# -*- coding: UTF-8 -*-
# Copyright © 2017 Cornelius Grotjahn - This file is covered by the LICENCE file in the project root directory.
import io
import os
import os.path
import re
import sqlite3
import sys
from typing import Optional, Iterable
import unittest

from .misc import set_indentation

SRC_DIR_PATH= os.path.dirname(sys.argv[0])

# [[[cog
# import cornelib
# make_struct= lambda cls,docstr: cornelib.cog_make_struct_type(cog.outl, cls, docstr)
#
# make_struct("Tabledef", '''The should-definition of a table.
#
# Args:
#   name (unicode): table name
#   statement_lines (list[unicode]): The create-statement split into lines.
#   column_types (dict[unicode, unicde]): column_name -> type for all columns in create-statement.
# ''')
#]]]
class Tabledef(object):
   # This is cog-generated code, make changes in the cog-comment above.
   @property
   def name(self):
      return self._name
   @property
   def statement_lines(self):
      return self._statement_lines
   @property
   def column_types(self):
      return self._column_types
   def __repr__(self):
      fields= ", ".join([
            "name="+repr(self._name),
            "statement_lines="+repr(self._statement_lines),
            "column_types="+repr(self._column_types),
         ])
      return "Tabledef({})".format(fields)
   # noinspection PyProtectedMember
   def __eq__(self, other):
      return (isinstance(other, Tabledef)
         and self._name == other._name
         and self._statement_lines == other._statement_lines
         and self._column_types == other._column_types
         )
   def __init__(self,
                name,
                statement_lines,
                column_types,
                ):
      """
      The should-definition of a table.

      Args:
        name (unicode): table name
        statement_lines (list[unicode]): The create-statement split into lines.
        column_types (dict[unicode, unicde]): column_name -> type for all columns in create-statement.
      """
      self._name= name
      self._statement_lines= statement_lines
      self._column_types= column_types
      assert isinstance(name, str)
      assert isinstance(statement_lines, list)
      assert isinstance(column_types, dict)
# [[[end]]]

def start_is_in(s, *beginnings):
   return any(s.lower().startswith(b+" ") for b in beginnings)

def read_clean_statements(script_path):
   """
   Returns iterable of SQL statements from given script file.

   :return: SQL lines forming one statement
   :rtype: str list
   """
   statement= []
   is_block= False
   with open(script_path) as input_stream:
      for line in input_stream:
         line= line.rstrip()
         stripped= re.sub(r"\s*(--.*)?$", "", line.strip())
         if stripped or statement:
            statement.append(line)
            if stripped=="begin":
               is_block= True
            if stripped.endswith(";") and (not is_block or stripped=="end;"):
               yield statement
               statement= []
               is_block= False
   if statement:
      raise Exception("incomplete imput stream, remaining code: "+" \n".join(statement))

def tabledef_from_statementlines(statement):
   first_line= statement[0]
   if not first_line.endswith("("):
      raise Exception("create table line does not end with '(': "+repr(first_line))
   table_name= first_line[:-1].split()[2]
   column_line_matches= (re.match(r"\s*(\w+)\s+(\w+)\W", line) for line in statement)
   column_types= { m.group(1):m.group(2)
                   for m in column_line_matches
                   if m and m.group(1).lower() not in{"create", "constraint"}
                 }
   return Tabledef(table_name, statement, column_types)

def isidentifier(s):
   return not s[0].isdigit() and all(c.isalnum() or c=="_" for c in s)

def read_new_schema(schema_script_path):
   """

   Returns: tuple of
     table_defs: Tabledef list
     trigger_defs: statement lines for triggers
     pass_through_defs: statement lines for other objects
   """
   trigger_defs= []
   pass_through_defs= []
   table_defs= []
   for statement in read_clean_statements(schema_script_path):
      first_line= statement[0]
      if start_is_in(first_line, "create table"):
         tabledef= tabledef_from_statementlines(statement)
         table_defs.append(tabledef)
      elif start_is_in(first_line, "create view", "create index", "create unique index"):
         pass_through_defs.append(statement)
      elif start_is_in(first_line, "create triggers"):
         trigger_defs.append(statement)
      else:
         raise Exception('cannot handle "{}"'.format(first_line))
   return table_defs, trigger_defs, pass_through_defs

def drop_programmability(db_conn: sqlite3.Connection) -> Iterable[str]:
   """removes all database objects except tables and returns list of the existing tables
   """
   rs= db_conn.execute("""select type, name
      from sqlite_master
      where not (type='index' and name like 'sqlite^_autoindex^_%' escape '^')""").fetchall()
   for typ, name in rs:
      if typ.lower() in {"trigger", "view", "index"}:
         db_conn.execute("drop {} {};".format(typ, name))
      elif typ.lower()=="table":
         yield name
      else:
         raise Exception("cannot handle {} object".format(typ))

def drop_constraints(existing_tables, new_tabledefs, db_conn: sqlite3.Connection) -> None:
   """
   drops all constraints from existing tables and creates all new columns while leaving all existing columns

   :param existing_tables: names of
   :type existing_tables: str set
   :param new_tabledefs: definitions of all tables to which is being upgrade
   :type new_tabledefs: Tabledef list
   """
   def make_tabledef(name, column_types):
      prev= "create table {}(".format(name)
      for colname,typ in column_types.items():
         yield prev
         prev= "  {} {},".format(colname,typ)
      yield prev.replace(",", ");")
   db_conn.execute("pragma foreign_keys= off;")
   new_column_types= { td.name:td.column_types for td in new_tabledefs }
   for table_name in existing_tables:
      existing_columns_result_set= db_conn.execute("pragma table_info('{}')".format(table_name)).fetchall()
      existing_column_names= [t[1] for t in existing_columns_result_set]
      old_and_new_column_types= {colname:coltype for (_i, colname, coltype, _notnull, _default, _pk) in existing_columns_result_set}
      if table_name in new_column_types:
         old_and_new_column_types.update(new_column_types[table_name])
      statement_lines= make_tabledef(table_name, old_and_new_column_types)
      hide_table(table_name, db_conn)
      recreate_table(table_name, statement_lines, existing_column_names, db_conn)
   for tabledef in new_tabledefs:
      if tabledef.name not in existing_tables:
         statement_lines= make_tabledef(tabledef.name, tabledef.column_types)
         db_conn.execute("\n".join(statement_lines))

def hidden_table_name(original_table_name):
   pfx= "_copy_of_"
   assert not original_table_name.startswith(pfx)
   return pfx+original_table_name

def hide_table(table_name: str, db_conn: sqlite3.Connection) -> None:
   db_conn.execute("alter table {} rename to {};".format(table_name, hidden_table_name(table_name)))

def recreate_table(table_name, statement_lines, copy_columns_list, db_conn: sqlite3.Connection) -> None:
   """
   Creates a table and copies existing data from hidden table.

   :param table_name: name of new table and indirectly of hidden table to be copied
   :type table_name: str
   :param statement_lines: SQL statement that creates the table
   :type statement_lines: str list
   :param copy_columns_list: columns that will be included in the data-copy
   :type copy_columns_list: str iterable
   """
   db_conn.execute("".join(statement_lines))
   columns_string= ", ".join(copy_columns_list)
   hidden= hidden_table_name(table_name)
   db_conn.execute("insert into {n}({l})\n  select {l}\n  from {h};".format(n=table_name, l=columns_string, h=hidden))
   db_conn.execute("drop table {};".format(hidden))

def create_constraints(table_defs, existing_tables, db_conn: sqlite3.Connection) -> None:
   new_table_names= set(t.name for t in table_defs)
   for table_name in new_table_names:
      hide_table(table_name, db_conn)
   for table_def in table_defs:
      recreate_table(table_def.name, table_def.statement_lines, table_def.column_types.keys(), db_conn)
   for t in set(existing_tables) - new_table_names:
      db_conn.execute("drop table {};".format(t))

def create_programmability(trigger_defs,
                           pass_through_defs,
                           db_conn: sqlite3.Connection) -> None:
   for statement_lines in pass_through_defs:
      db_conn.execute("\n".join(statement_lines))
   for i, statement_lines in enumerate(trigger_defs):
      _create, _triggers, table_name, columns = statement_lines[0].split(None, 3)
      core_statement= "\n      ".join(statement_lines[1:])
      for insert_update in ("insert", "update of "+columns):
         sql= set_indentation(0, """
            create trigger {table}_{i}{iu}
               before {insert_update} on {table}
               for each row
            {statement}""")
         db_conn.execute(sql.format(i= i,
                                 iu= insert_update[0],
                                 insert_update= insert_update,
                                 table= table_name,
                                 statement= core_statement,
                                 ))
      # getting trigger to check existing data by creating it on dummy table and performing dummmy insert:
      db_conn.execute("create table tmpt(x int);")
      core_statement= re.sub(r"\b(OLD|NEW)\b", table_name, core_statement)
      core_statement= re.sub(r"\bfrom\b", "from {},".format(table_name), core_statement)
      sql= set_indentation(0, """
         create trigger tmptr
            before insert on tmpt
            for each row
         {}""".format(core_statement))
      db_conn.execute(sql)
      db_conn.execute("insert into tmpt values(0);")
      db_conn.execute("drop table tmpt;")

class CommandLineData(object):
   """
   Singleton representing upgrade database and input scripts.
   Members:
      database_path                 path of sqlite3 database to be upgraded
      upgrades_to_version           version to which upgrade scripts take us
      schema_script_path            path of file containing schema definition
      upgrade_script_paths[from_version] paths if upgrade script file that increment `from_version` to upgrades_to_version.
   """
   def __init__(self):
      args= sys.argv[1:]
      if len(args)<2:
         raise Exception("Expected at least 2 arguments")
      self.database_path= args.pop(0)
      if not self.database_path.endswith(".sl3"):
         raise Exception("Expected .sl3 files as first argument, not "+self.database_path)
      self.schema_script_path= args.pop(0)
      if not self.schema_script_path.endswith(".sql"):
         raise Exception("Expected schema sql file as second argument, not "+self.schema_script_path)

      self.upgrade_script_paths= {}
      for p in args:
         splut_filename= os.path.basename(p).split(".")
         if len(splut_filename)==3 and splut_filename[0]=="upgrade" and splut_filename[2]=="sql":
            upgrade_script_version= int(splut_filename[1])
            self.upgrade_script_paths[upgrade_script_version]= p
         else:
            raise Exception("Expected upgrade.*.sql file, not "+p)
      self.upgrades_to_version= len(self.upgrade_script_paths)
      if sorted(self.upgrade_script_paths) != range(self.upgrades_to_version):
         raise Exception("Expected upgrade scripts 'upgrade.<n>.sql' with consecutive <n> starting at 0")


def script_main():
   # TODO re-implement using schapp_upgrade
   commandline_data= CommandLineData()
   table_defs, trigger_defs, pass_through_defs= read_new_schema(commandline_data.schema_script_path)
   version_table= tabledef_from_statementlines([
      "create table _database_version(",
      "  current_version int not null,",
      "  constraint only_one_row check(rowid==1)"
      ");"])
   table_defs.append(version_table)
   with sqlite3.connect(commandline_data.database_path, isolation_level=None) as conn:
      cursor= conn.cursor()
      class EchoingConnection(sqlite3.Connection):
         def execute(sql, params=None):
            print("sql>",sql)
            return conn.execute(sql, params)
      db_conn= EchoingConnection()
      existing_tables= set(drop_programmability(db_conn))
      drop_constraints(existing_tables, table_defs, db_conn)
      if not existing_tables:
         # _database_version was not on DB (DB created from scratch) and is not in existing_tables
         # but it exists now because drop_constraints creates all new columns.
         db_conn.execute("insert into _database_version(current_version) values(0);")
         existing_tables= {"_database_version"}
         current_version= 0
      else:
         current_version_rs= db_conn.execute("select current_version from _database_version").fetchall()
         assert len(current_version_rs)==1
         assert len(current_version_rs[0])==1
         current_version= int(current_version_rs[0][0])

      # Running the appropriate upgrade script to upgrade FROM the current version to head.
      if current_version > commandline_data.upgrades_to_version:
         raise Exception("DB is too new (v{} > v{})".format(current_version, commandline_data.upgrades_to_version))
      if 0 < current_version < commandline_data.upgrades_to_version:
         errors= False
         for stmt_lines in read_clean_statements(commandline_data.upgrade_script_paths[current_version]):
            for row in db_conn.execute(*stmt_lines):
               print(row)
               errors= True
         if errors: raise Exception("there was integrity checking output")

      create_constraints(table_defs, existing_tables, db_conn)
      create_programmability(trigger_defs, pass_through_defs, db_conn)
      db_conn.execute("pragma foreign_keys= on;")
      violations= db_conn.execute("pragma foreign_key_check;").fetchall()
      if violations:
         raise Exception("FK violations: " + "\n  ".join(repr(row) for row in violations))
      db_conn.execute("update _database_version set current_version= ?", [commandline_data.upgrades_to_version])

def schapp_upgrade(db_conn: sqlite3.Connection,
                   target_version: str,
                   schema_script_path: str,
                   upgrade_script_path: Optional[str]) -> None:
   """
   Changes a database to given schema:

   - Creates all new column from schema, leaving all existing ones (even if not in new schema)
   - Drops all constraints
   - Runs given upgrade script which is expected to populate new columns and clear old ones.
   - Drops all old columns, checking they are now empty.
   - Creates all constraints, views, and triggers.

   :param db_conn: the database to be upgraded
   :param target_version: version after upgrade, to be written to _database_version table
   :param schema_script_path: file containing target schema
   :param upgrade_script_path: file containing data conversion statements from current version to target_version
                               or None if there is no upgrade code
   """
   table_defs, trigger_defs, pass_through_defs= read_new_schema(schema_script_path)
   version_table= tabledef_from_statementlines([
      "create table _database_version(",
      "  current_version text not null,",
      "  constraint only_one_row check(rowid==1)"
      ");"])
   table_defs.append(version_table)
   existing_tables= set(drop_programmability(db_conn))
   drop_constraints(existing_tables, table_defs, db_conn)

   # Running the appropriate upgrade script to upgrade FROM the current version to head.
   if upgrade_script_path is not None:
      errors= False
      for stmt_lines in read_clean_statements(upgrade_script_path):
         for row in db_conn.execute("\n".join(stmt_lines)):
            print(row)
            errors= True
      if errors: raise Exception("there was integrity checking output")

   create_constraints(table_defs, existing_tables, db_conn)
   create_programmability(trigger_defs, pass_through_defs, db_conn)
   db_conn.execute("pragma foreign_keys= on;")
   violations= db_conn.execute("pragma foreign_key_check;").fetchall()
   if violations:
      raise Exception("FK violations: " + "\n  ".join(repr(row) for row in violations))
   if db_conn.execute("select * from _database_version").fetchall():
      db_conn.execute("update _database_version set current_version= ?", [target_version])
   else:
      db_conn.execute("insert into _database_version select ?", [target_version])

class _Test_Schapp_(unittest.TestCase):
   def test_tabledefs_ending_comma(self):
      script= set_indentation(0, """
         create table T(
            hello int,
            world text not null);""")
      tabledefs,triggers,other = read_new_schema(script)
      self.assertFalse(triggers)
      self.assertFalse(other)
      self.assertEqual(len(tabledefs), 1)
      t= tabledefs[0]
      self.assertEqual(t.name, "T")
      self.assertEqual(t.statement_lines, script.split("\n"))
      self.assertEqual(t.column_types, {'hello':'int', 'world':'text'})

if __name__=="__main__":
   script_main()



# -*- coding: UTF-8 -*-
# Copyright © 2017 Cornelius Grotjahn - This file is covered by the LICENCE file in the project root directory.
import os
import re

def cog_make_struct_type(output_function, class_name, docstring):
   """
   Outputs code for immutable "struct" class. Result is without trailing newline.

   Include comments with calls to cog_make_struct_type() in some source file. Then run
   "cog.py -r source.py" over it. Cog can be installed with "pip install cogapp".

   Example:
      Only first occurrence needs the import and may define a function with a default for output_function::

      # [[[cog
      # import cornelib
      # make_struct= lambda cls,docstr: cornelib.cog_make_struct_type(cog.outl, cls, docstr)
      #
      # make_struct("Coord", '''
      # A two-dimensional coordinate.
      #
      # Args:
      #   x (float): x-value
      #   y (float): y-value counting up from bottom
      # ''')
      # ]]]
      class Coord(object):
         # This is cog-generated code, make changes in the cog-comment above.
         ...
      # [[[end]]]

   Args:
      class_name (unicode): name for new class
      docstring (unicode): for __init__ of new class.
         Lines in docstring starting "name (type):" define the class's fields.
         Indentation is irrelevant but must be consistent for all lines including the first one.
         May lead with a newline.
      output_function (lambda): function used to output one line (without newline), typically `cog.outl`
   """
   raw_code= r'''
      class CLASSNAME(object):
         # This is cog-generated code, make changes in the cog-comment above.
         @property               \
         def FIELDNAME(self):    \
            return self._FIELDNAME
         def __repr__(self):
            fields= ", ".join([
                  "FIELDNAME="+repr(self._FIELDNAME),
               ])
            return "CLASSNAME({})".format(fields)
         # noinspection PyProtectedMember
         def __eq__(self, other):
            return (isinstance(other, type(self))
               and self._FIELDNAME == other._FIELDNAME
               )
         def __init__(self,
                      FIELDNAME,
                      ):
            """
            DOCSTRING
            """
            self._FIELDNAME= FIELDNAME
            assert isinstance(FIELDNAME, FIELDTYPE)
      '''
   raw_code= set_indentation(0, raw_code)
   raw_code= raw_code.replace("CLASSNAME", class_name)
   raw_code= raw_code.replace("      DOCSTRING", set_indentation(6, docstring))
   fielddef= re.compile(r"^\s*(\w+)\s+\((\w+)(\[[^)]*)?\):")
   fielddef_matches= [fielddef.match(line) for line in docstring.split("\n")]
   fields= [ (m.group(1), m.group(2)) for m in fielddef_matches if m]
   carry= None
   for code_line in raw_code.split('\n'):
      if carry:
         code_line= carry +'\n' +code_line
         carry= None
      if code_line.endswith('\\'):
         carry= code_line[:-1].rstrip()
         continue
      code_line= code_line.rstrip()
      if "FIELDNAME" in code_line:
         for field_name, field_type in fields:
            output_function(code_line.replace("FIELDNAME", field_name).replace("FIELDTYPE", field_type))
      else:
         output_function(code_line)

def set_indentation(new_indentation, multi_line_string):
   """
   returns multi_line_string with
      - all trailing spaces stripped
      - all leading and trailing empty lines stripped
      - no final newline
      - smallest indentation set to `new_indentation`
   """
   assert '\t' not in multi_line_string
   lines= [l.rstrip() for l in multi_line_string.split('\n')]
   while not lines[0]:
      lines.pop(0)
   while not lines[-1]:
      lines.pop()
   spaces= re.compile("\s*")
   old_indentation= min(spaces.match(l).end() for l in lines if l)
   if new_indentation > old_indentation:
      pfx= ' ' * (new_indentation - old_indentation)
      lines= [pfx + l if l else "" for l in lines]
   elif new_indentation < old_indentation:
      cut= old_indentation - new_indentation
      lines= [l[cut:] for l in lines]
   return '\n'.join(lines).rstrip()

class partial_write_guard(object):
   def __init__(self, path:str):
      """
      Context manager protecting against leaving partially written files in case of error.
      Write to the path returned when entering; this will then be renamed to the given path,
      or deleted.

      :param path: the intended path for a file
      """
      self.path= path
      self.partial= path+".part"
   def __enter__(self):
      return self.partial
   def __exit__(self, exc_type, exc_val, exc_tb):
      if exc_type is None:
         os.rename(self.partial, self.path)
      else:
         os.unlink(self.partial)



# -*- coding: UTF-8 -*-
# Copyright © 2017 Cornelius Grotjahn - This file is covered by the LICENCE file in the project root directory.
import base64
import logging
import lzma
import os

from cryptography.hazmat.primitives import ciphers, hmac, hashes
from cryptography.hazmat.primitives.kdf import pbkdf2
from cryptography.hazmat import backends

def _password_derive(key, salt):
   kdf = pbkdf2.PBKDF2HMAC(
      algorithm=hashes.SHA256(),
      length=16,
      salt=salt,
      iterations=100000,
      backend=backends.default_backend()
   )
   return base64.urlsafe_b64encode(kdf.derive(key))

class _WriteableEncryptedStream(object):
   """A write-only stream to an encrypted file. Data written to the stream is automatically encrypted."""
   def __init__(self, path, password):
      backend = backends.default_backend()
      str_key = password.encode()
      salt = os.urandom(16)
      hmac_salt = os.urandom(16)
      iv = os.urandom(16)
      key = _password_derive(str_key, salt)
      hmac_key = _password_derive(str_key, hmac_salt)
      self.checksummer = hmac.HMAC(hmac_key, hashes.SHA256(), backend=backend)
      cipher = ciphers.Cipher(ciphers.algorithms.AES(key), ciphers.modes.CTR(iv), backend=backend)
      self.cryptor = cipher.encryptor()
      initialiser= iv + salt + hmac_salt
      self.checksummer.update(initialiser)
      self.output_stream= open(path, mode="wb")
      self.output_stream.write(initialiser)
   def write(self, data):
      out = self.cryptor.update(data)
      self.checksummer.update(out)
      self.output_stream.write(out)
   def close(self):
      out = self.cryptor.finalize()
      self.checksummer.update(out)
      self.output_stream.write(out)
      signature = self.checksummer.finalize()
      self.output_stream.write(signature)
      self.output_stream.close()

class _ReadableEncryptedStream(object):
   """A read-only stream to an encrypted file. Data read from the stream is automatically decrypted."""
   def __init__(self, path, password):
      self.input_stream= open(path, "rb")
      str_key = password.encode()
      raw= self.input_stream.read(80)
      if len(raw)<80:
         raise RuntimeError("file too short")
      assert len(raw)==80
      initialiser= raw[:48]
      iv = initialiser[:16]
      salt = initialiser[16:32]
      hmac_salt = initialiser[32:48]
      self.read_ahead= raw[48:] # empty indicates end of input_stream
      self.outbuffer= b""
      key = _password_derive(str_key, salt)
      hmac_key = _password_derive(str_key, hmac_salt)
      backend = backends.default_backend()
      cipher = ciphers.Cipher(ciphers.algorithms.AES(key), ciphers.modes.CTR(iv), backend=backend)
      self.cryptor = cipher.decryptor()
      self.checksummer = hmac.HMAC(hmac_key, hashes.SHA256(), backend=backend)
      self.checksummer.update(initialiser)
   def _fill(self, size):
      if not self.read_ahead:
         return
      bs= 8192 if size<0 else size
      while size==-1 or size>len(self.outbuffer):
         raw= self.read_ahead + self.input_stream.read(bs)
         encrypted= raw[:-32]
         self.read_ahead= raw[-32:]
         if len(encrypted) > 0:
            self.outbuffer += self.cryptor.update(encrypted)
            self.checksummer.update(encrypted)
         else:
            # reached the end which means read_ahead is signature
            self.outbuffer += self.cryptor.finalize()
            self.checksummer.verify(self.read_ahead)
            self.read_ahead= None
            break
   def read(self, size=-1):
      self._fill(size)
      n= len(self.outbuffer) if size<0 or len(self.outbuffer)<size else size
      data= self.outbuffer[:size]
      self.outbuffer= self.outbuffer[size:]
      return data
   def seekable(self):
      return False
   def close(self):
      self.input_stream.close()

class open_bin(object):
   def __init__(self, path, mode, password):
      """Opens an encrypted binary stream in a context manager."""
      if mode == "w":
         fd= _WriteableEncryptedStream(path, password)
      elif mode == "r":
         fd= _ReadableEncryptedStream(path, password)
      else:
         raise ValueError("Invalid mode: "+mode)
      # Derived classes may append more streams using fd as underlying stream.
      self._streams= [fd]
   def __enter__(self):
      return self._streams[-1]
   def __exit__(self, exc_type, exc_val, exc_tb):
      while self._streams:
         fd= self._streams.pop()
         fd.close()

class open_txt(open_bin):
   def __init__(self, path, mode, password):
      """Opens an encrypted and compressed UTF-8 stream."""
      open_bin.__init__(self, path, mode, password)
      f= lzma.open(self._streams[-1], mode+"t", encoding="UTF-8")
      self._streams.append(f)

class open_lzma(open_bin):
   def __init__(self, path, mode, password):
      """Opens an encrypted and compressed binary stream."""
      open_bin.__init__(self, path, mode, password)
      f= lzma.open(self._streams[-1], mode+"b")
      self._streams.append(f)



# -*- coding: UTF-8 -*-
# Copyright © 2017 Cornelius Grotjahn - This file is covered by the LICENCE file in the project root directory.
import unittest
from .misc import set_indentation, cog_make_struct_type

class _Cornelib_Tests_(unittest.TestCase):
   def test_set_indentation(self):
      self.assertEqual(set_indentation(0, "\n  two \n   three\n"), "two\n three")
      self.assertEqual(set_indentation(0, "\n  two \n\n   three\n"), "two\n\n three")
   def test_cog_make_struct_type(self):
      docstring= """
               Defines the way to update a single database column when data is being saved. Any names included in braces
               denote display column names and are replaced with values from changed display data.

               Args:
                  where_clause (unicode): clause following where keyword that defines row to be updated (from display columns)
                  uses_grid_columns (frozenset[unicode]): list of display column used in set_clause (used for determining quickly whether
                     this UpdateColumnDef needs to be included in an updated, given changed display columns.
               """
      expected= set_indentation(0, '''
         class UpdateColumnDef(object):
            # This is cog-generated code, make changes in the cog-comment above.
            @property
            def where_clause(self):
               return self._where_clause
            @property
            def uses_grid_columns(self):
               return self._uses_grid_columns
            def __repr__(self):
               fields= ", ".join([
                     "where_clause="+repr(self._where_clause),
                     "uses_grid_columns="+repr(self._uses_grid_columns),
                  ])
               return "UpdateColumnDef({})".format(fields)
            # noinspection PyProtectedMember
            def __eq__(self, other):
               return (isinstance(other, type(self))
                  and self._where_clause == other._where_clause
                  and self._uses_grid_columns == other._uses_grid_columns
                  )
            def __init__(self,
                         where_clause,
                         uses_grid_columns,
                         ):
               """''' + docstring+ '''"""
               self._where_clause= where_clause
               self._uses_grid_columns= uses_grid_columns
               assert isinstance(where_clause, unicode)
               assert isinstance(uses_grid_columns, frozenset)''')
      result= []
      cog_make_struct_type(lambda l: result.append(l), "UpdateColumnDef", docstring)
      actual= '\n'.join(result)
      self.maxDiff= None
      self.assertEqual(actual, expected)

if __name__=="__main__":
   unittest.main()



# -*- coding: UTF-8 -*-
# Copyright © 2017 Cornelius Grotjahn - This file is covered by the LICENCE file in the project root directory.
import itertools
import logging
import os
import shutil
import sqlite3
import time
import unittest

from abus import database
from abus import main

class AbusTestBase(unittest.TestCase):
   def change_latest_runname(self, offset):
      """
      Adds `offset` to the latest run_name to simulate backup happening earlier.

      :type db: database.Database
      :param offset: seconds to be added to all timestamps
      """
      with database.connect(self.databasepath, self.archivedir) as db:
         with db._get_connection() as conn:
            latest, archive_dir= conn.execute("""
               select run_name, archive_dir
               from run
               order by run_name desc
               limit 1""").fetchall()[0]
            new_name= db._make_runname(time.mktime(time.strptime(latest, "%Y_%m_%d_%H%M"))+offset)
            for table in "content", "run":
               conn.execute("update "+table+" set run_name= ? where run_name = ?", [new_name, latest])

      # renaming index file and content file
      for relpath in archive_dir +"/" +latest +".lst", latest +".gz":
         path= self.archivedir + "/" +relpath
         os.rename(path, path.replace(latest, new_name))

   def create_homedir_file(self, relative_path, prime, timestamp=None):
      path= self.homedir+"/"+relative_path
      os.makedirs(os.path.dirname(path), exist_ok=True)
      with open(path,"wb") as f:
         f.write(bytes(n*prime%256 for n in range(8192)))
      if timestamp:
         os.utime(path, times=(timestamp,timestamp))

   @property
   def databasepath(self):
      return self.archivedir+"/index.sl3"

   def executesql(self, sql):
      """Runs sql statement on the index database and returns all rows.
      """
      with self.get_direct_db_connection() as conn:
         return conn.execute(sql).fetchall()

   def find_files(self, start):
      """Returns direntries for all files in `start`, recursively"""
      q= [start]
      while q:
         entries= list(os.scandir(q.pop())) # prevents iterator leakage
         yield from (e for e in entries if e.is_file())
         q.extend(e.path for e in entries if e.is_dir())

   def get_direct_db_connection(self):
      conn= sqlite3.connect(self.databasepath)
      conn.isolation_level= None # autocommit
      conn.execute("PRAGMA synchronous=OFF")
      return conn

   def remove_dir_contents(self, path):
      for direntry in os.scandir(path):
         if direntry.is_dir():
            shutil.rmtree(direntry.path)
         elif direntry.is_file():
            os.unlink(direntry.path)

   def setup_directories(self):
      """
      Sets up empty home, archive, and restore dirs, config file and empty database.

      :rtype: None
      """
      logging.basicConfig(level=logging.DEBUG)
      candidates_for_tmp= ["C:/tmp", "C:/temp"]
      if "TEMP" in os.environ: candidates_for_tmp.append(os.environ["TEMP"])
      for tempdir in candidates_for_tmp:
         if os.path.isdir(tempdir):
            break
      else:
         raise Exception("Could not find a temporary directory for tests")

      self.password= "Sesam, öffne dich!"
      self.homedir= tempdir + "/abushome"
      self.otherdir= tempdir + "/abusother"
      self.archivedir= tempdir + "/abusarchive"
      self.restoredir= tempdir + "/abusrestore"
      os.makedirs(self.homedir, exist_ok=True)
      os.makedirs(self.archivedir, exist_ok=True)
      os.makedirs(self.restoredir, exist_ok=True)
      os.makedirs(self.otherdir, exist_ok=True)
      os.chdir(self.restoredir)

      # empty them
      self.remove_dir_contents(self.homedir)
      self.remove_dir_contents(self.archivedir)
      self.remove_dir_contents(self.restoredir)
      self.remove_dir_contents(self.otherdir)

      self.configfile= self.otherdir + "/abusconfig.cfg"
      self.write_config_file()

      with database.connect(self.databasepath, self.archivedir, allow_create=True):
         pass

   def write_config_file(self, *extra_param_lines):
      with open(self.configfile, "w", encoding="utf-8") as f:
         print("logfile", self.otherdir + "/abuslog.txt", file=f)
         print("archive", self.archivedir, file=f)
         print("password", self.password, file=f)
         print("\n".join(extra_param_lines), file=f)
         print("[include]", file=f)
         print(self.homedir, file=f)

   def setup_backup_with_well_known_checksums(self):
      self.setup_directories()
      self.expected_backups= set() # contains (path rel to homedir, backup filename) pairs

      path= "my_valueable_file"
      with open(self.homedir+"/"+path, "w", encoding="utf-8") as f:
         for n in range(400):
            for p in range(10):
               print(n**p, file=f, end="")
            print(file=f)
      self.expected_backups.add((path,"73672933c00ab3cd730c8715a392ee6dee9ba2c0a8e5e5d07170b6544b0113ef.z"))

      path= "my_little_file"
      with open(self.homedir+"/"+path, "wb") as f:
         f.write(bytearray(range(256)))
      self.expected_backups.add((path,"40aff2e9d2d8922e47afd4648e6967497158785fbd1da870e7110266bf944880"))

      path= "subdir_a/I_am_in_a.tif"
      self.create_homedir_file(path, 31)
      self.expected_backups.add((path,"498630374d56ea4c53c5baaee00b630be09ca5dbe678b11b1a81b37f1058635f"))

      path= "subdir_b/I_am_in_b.bin"
      self.create_homedir_file(path, 97)
      self.expected_backups.add((path,"1244d0711c6534ee79a6d3b82cea33b76e766e46cbca75791ac9fa3a30e365f3.z"))

      rc= main.main(["test", "-f", self.configfile, "--backup"])
      self.assertEqual(rc, 0)

   def setup_multiple_backups(self):
      self.setup_directories()
      def primes():
         prev= [2]
         for c in itertools.count(3,2):
            for p in prev:
               if p*p>c:
                  yield c
                  prev.append(c)
                  break
               if c%p==0: break
      primes= primes().__iter__()

      all_paths= ["a", "b", "s/a", "s/b", "t/a", "t/b"]
      n= len(all_paths)*3//4
      for i in range(3):
         change_paths= all_paths if i==0 else all_paths[:n] if i==1 else all_paths[-n:]
         offset= (i-2)*86000 -3000
         timestamp= time.time() + offset
         for path in change_paths:
            self.create_homedir_file(path, next(primes), timestamp)
         rc= main.main(["test", "-f", self.configfile, "--backup"])
         self.assertEqual(rc, 0)
         if i<2:
            self.change_latest_runname(offset+1000)



# -*- coding: UTF-8 -*-
# Copyright © 2017 Cornelius Grotjahn - This file is covered by the LICENCE file in the project root directory.
import gzip
import logging
import os
import shutil
import sqlite3
import unittest.mock
from abus import backup
from abus import config
from abus import database
from abus import crypto
from abus import main
from abus import purge
from abus.testbase import AbusTestBase

class BackedUpHomedirTests(AbusTestBase):
   def test_simple_end_to_end_backup(self):
      self.setup_backup_with_well_known_checksums()

      # checking archive contents matches DB
      subdir,run_name = self.executesql("select archive_dir, run_name from run")[0]
      index_file_path_rel= subdir+"/"+run_name+".lst"
      expected_backupfiles= {"index.sl3", index_file_path_rel, run_name+".gz"}
      entries= self.executesql("""
         select path, archive_dir, location.checksum, is_compressed
         from content
            join location on location.checksum = content.checksum""")
      expected_backupfiles.update(a+"/"+c+(".z" if z else "") for p,a,c,z in entries)
      expected_backupfiles_abs= {self.archivedir+"/"+rel for rel in expected_backupfiles}
      actual= set(direntry.path.replace("\\","/")
                  for direntry in self.find_files(self.archivedir))
      self.assertEqual(actual, expected_backupfiles_abs)

      # checking archive contents matches what has been backed up
      # rs= e.g. ('C:/tmp/abushome/my_little_file', '40aff2e9d2d8922e47afd4648e6967497158785fbd1da870e7110266bf944880', 0)
      from_db= {(p, c+".z" if z else c) for p,a,c,z in entries}
      from_test= {(self.homedir+"/"+p, f) for p,f in self.expected_backups}
      self.assertEqual(from_db, from_test)

      # checking that index files matches
      with crypto.open_txt(self.archivedir+"/"+index_file_path_rel, "r", self.password) as f:
         index_entries= set(l.strip() for l in f)
      dbentries= set("{} {} {}".format(*row)
                     for row in self.executesql("select checksum, timestamp, path from content"))
      self.assertEqual(index_entries, dbentries)

      # checking that contents file matches
      with gzip.open(self.archivedir+"/"+run_name+".gz", "rt", encoding="UTF-8") as stream:
         lines= {l.strip() for l in stream}
      self.assertEqual(lines, expected_backupfiles)

      # decrypting backup
      my_litte_backup= [e for e in expected_backupfiles_abs if "/40aff2e" in e]
      self.assertEqual(len(my_litte_backup), 1)
      self.assertEqual(os.stat(my_litte_backup[0]).st_size, 256+80) # 80 for salt, init vector, and checksum
      with crypto.open_bin(my_litte_backup[0], "r", self.password) as f:
         blk= f.read(8192)
         self.assertEqual(len(blk), 256)
         self.assertEqual(blk, bytearray(range(256)))

   def test_error_prevents_closing_backup_run(self):
      self.setup_directories()
      self.create_homedir_file("unreadable",23)
      with unittest.mock.patch('abus.backup.make_backup_copy', side_effect=RuntimeError):
         rc= main.main(["test", "-f", self.configfile, "--backup"])
      self.assertEqual(rc, 1)
      rs= self.executesql("select run_name, archive_dir, is_complete from run")
      self.assertEqual(len(rs), 1)
      run_name, archive_dir, is_complete = rs[0]
      self.assertEqual(is_complete, 0)
      index_file_path= self.archivedir+"/"+rs[0][1]+"/"+rs[0][0]+".lst"
      with crypto.open_txt(index_file_path, "r", self.password) as stream:
         contents= set(stream)
      error_line= "error 0 {}/unreadable\n".format(self.homedir)
      self.assertIn(error_line, contents)
      self.assertTrue(os.path.isfile(self.archivedir+"/"+run_name+".gz"))

   def test_changing_file_gets_backed_up_after_retry(self):
      self.setup_directories()
      self.create_homedir_file("myfile",23)
      def make_backup_copy_sideeffect(path, expected_checksum, backup_path, open_dst_function, password):
         self.create_homedir_file("myfile",97) # 1244d0711c6534ee79a6d3b82cea33b76e766e46cbca75791ac9fa3a30e365f3
      with unittest.mock.patch('time.sleep'):
         with unittest.mock.patch('abus.backup.make_backup_copy',
                                  wraps=backup.make_backup_copy,
                                  side_effect=make_backup_copy_sideeffect):
            rc= main.main(["test", "-f", self.configfile, "--backup"])
      self.assertEqual(rc, 0)
      rs= self.executesql("select run_name, archive_dir, is_complete from run")
      self.assertEqual(len(rs), 1)
      _run_name, _archive_dir, is_complete = rs[0]
      self.assertEqual(is_complete, 1)
      rs= self.executesql("""select location.checksum
         from content join location on location.checksum = content.checksum""")
      self.assertEqual(len(rs), 1)
      self.assertEqual(rs[0][0], "1244d0711c6534ee79a6d3b82cea33b76e766e46cbca75791ac9fa3a30e365f3")

   def test_always_changing_file_causes_error_after_5_attempts(self):
      self.setup_directories()
      primes= iter([17,19,23,29,31,37,41])
      self.create_homedir_file("myfile", next(primes))
      def make_backup_copy_sideeffect(path, expected_checksum, backup_path, open_dst_function, password):
         self.create_homedir_file("myfile", next(primes))
      with unittest.mock.patch('time.sleep'):
         with unittest.mock.patch('abus.backup.make_backup_copy',
                                  wraps=backup.make_backup_copy,
                                  side_effect=make_backup_copy_sideeffect) as mock:
            rc= main.main(["test", "-f", self.configfile, "--backup"])
            self.assertEqual(mock.call_count, 5)
      self.assertEqual(rc, 1)
      rs= self.executesql("select is_complete from run")
      self.assertEqual(rs, [(0,)])
      rs= self.executesql("select * from location")
      self.assertEqual(rs, [])

   def test_bad_include_entry_does_not_stop_other_backups(self):
      """bug: read error on c:/users/*/documents causes unhandled exception and aborts whole backup"""
      self.setup_directories()
      # Appending bad [include]-entry happens to run it first and used to prevent homedir from being backed up
      with open(self.configfile, "a") as stream:
         stream.write("C:/Users/*/Desktop\n")
      self.create_homedir_file("a_file",23)
      rc= main.main(["test", "-f", self.configfile, "--backup"])
      self.assertEqual(rc, 1)
      with self.get_direct_db_connection() as conn:
         rows= [conn.execute("select * from "+t).fetchall() for t in("content", "run", "location", "run where is_complete")]
      self.assertEqual(len(rows[0]), 1)
      self.assertEqual(len(rows[1]), 1)
      self.assertEqual(len(rows[2]), 1)
      self.assertEqual(len(rows[3]), 0)

   def test_archive_dirs_to_use(self):
      def go(initial_usage: dict, expected_result: list):
         i= iter(backup.archive_dirs_to_use(initial_usage))
         for archive_dir, n, skip in expected_result:
            actual= [next(i) for _ in range(n)]
            expect= [archive_dir]*n
            self.assertEqual(actual, expect)
            for _ in range(skip):
               next(i)
      go({},
         [("00", 100, 0), ("01", 100, 9800), ("00/01", 100, 0), ("01/01", 100, 9800), ("00/02", 100, 979800),
          ("99/99", 100, 0),
          ("00/00/01", 100, 0)])
      go({"00": 100, "45": 30},
         [("45", 70, 0), ("01", 100, 9700), ("00/01", 100, 0), ("01/01", 100, 9800), ("00/02", 100, 0)])

   def test_indexdb_in_different_location(self):
      self.setup_directories()
      self.create_homedir_file("hello.txt", 79)
      db_path= self.otherdir+"/index_database"
      shutil.move(self.databasepath, db_path)
      self.write_config_file("indexdb " +db_path)
      rc= main.main(["test", "-f", self.configfile, "--backup"])
      self.assertEqual(rc, 0)
      with sqlite3.connect(db_path) as conn:
         content= conn.execute("select * from content").fetchall()
         subdir,run_name = conn.execute("select archive_dir, run_name from run").fetchall()[0]
      self.assertEqual(len(content), 1)
      self.assertFalse(os.path.exists(self.databasepath))

      content_file_path= self.archivedir+"/"+run_name+".gz"
      with gzip.open(content_file_path, "rt", encoding="UTF-8") as stream:
         lines= {l.strip() for l in stream}
      lines.remove(run_name+".gz")
      lines.remove(subdir+"/"+run_name+".lst")
      self.assertEqual(len(lines), 1)
      self.assertRegex(lines.pop(), "[a-z0-9]{64}")

      # bug: Exception in os.path.relpath if db on a different drive.
      os.unlink(content_file_path)
      cfg= config.Configuration(["test", "-f", self.configfile, "--backup"])
      with database.connect(cfg.database_path, cfg.archive_root_path, allow_create=False) as db:
         cfg.database_path= "X:/indexdb" # database_path no linger used, we can lie about it now
         with self.assertRaises(ValueError):
            os.path.relpath(cfg.database_path, cfg.archive_root_path)
         purge.write_content_file(cfg, db)

   def test_db_u64_ino_dev(self):
      self.setup_directories()
      with database.connect(self.databasepath, self.archivedir) as db:
         db.remember_checksum(2**63, 2**63, 1500000000.0, '1244d0711c6534ee79a6d3b82cea33b76e766e46cbca75791ac9fa3a30e365f3')
         checksum= db.lookup_checksum(2**63, 2**63, 1500000000.0)
      self.assertEqual(checksum, '1244d0711c6534ee79a6d3b82cea33b76e766e46cbca75791ac9fa3a30e365f3')



# -*- coding: UTF-8 -*-
# Copyright © 2017 Cornelius Grotjahn - This file is covered by the LICENCE file in the project root directory.
import os
import time
import unittest
from abus import config
from abus import crypto
from abus import database
from abus import main
from abus.testbase import AbusTestBase

class PurgeTests(AbusTestBase):
   def test_purge_e2e(self):
      self.setup_directories()
      def checkdb(n_runs):
         data= self.executesql("""
            select run_name, content.checksum, location.checksum
            from content
               left join location on location.checksum = content.checksum""")
         self.assertEqual(len(set(r for r,c,l in data)), n_runs)
         self.assertEqual(len(data), n_runs*2)
         # test always keeps 2 files current:
         self.assertEqual(len(set(l for r,c,l in data if l is not None)), 2)
      def contents_runs(expected_content_size, expected_runs_size):
         """Checks that there are as many backup files and runs in archivedir as expected

         :return: Pair with sets with names of backup files and run files respectively.
         """
         all_files= set(direntry.name for direntry in self.find_files(self.archivedir))
         runs= set(n for n in all_files if n.endswith(".lst"))
         content_file= max(runs).replace(".lst", ".gz")
         backups= all_files -runs -{"index.sl3", content_file}
         self.assertEqual(len(backups), expected_content_size)
         self.assertEqual(len(runs), expected_runs_size)
         return backups, runs
      def elt(s):
         """returns the only element from set"""
         self.assertEqual(len(s), 1)
         for e in s: return e
      self.create_homedir_file("file_a", 23, time.time()-365*86400)
      self.create_homedir_file("file_b.gz", 11, time.time()-365*86400)
      rc= main.main(["test", "-f", self.configfile, "--backup"])
      self.assertEqual(rc, 0)
      print("cornie",45)
      self.change_latest_runname(-365*86400)
      content, runs= contents_runs(2, 1) # file_a0, file_b0, run0
      compressed= {n for n in content if n.endswith(".z")}
      file_a0= elt(compressed)
      file_b0= elt(content - compressed)
      run0= elt(runs)
      checkdb(n_runs=1)

      self.create_homedir_file("file_a", 17, time.time()-335*86400)
      rc= main.main(["test", "-f", self.configfile, "--backup"])
      self.assertEqual(rc, 0)
      self.change_latest_runname(-335*86400)
      # file_a0 has been purged
      content, runs= contents_runs(2, 2) # file_a1, file_b0, run0, run1
      file_a1= elt(content - {file_b0})
      run1= elt(runs - {run0})
      checkdb(n_runs=2)

      self.create_homedir_file("file_b.gz", 59, time.time()-305*86400)
      rc= main.main(["test", "-f", self.configfile, "--backup"])
      self.assertEqual(rc, 0)
      self.change_latest_runname(-305*86400)
      content, runs= contents_runs(2, 2) # file_a1, file_b1, run1, run2
      file_b1= elt(content - {file_a1})
      run2= elt(runs - {run1})
      checkdb(n_runs=2)

   def test_db_get_purgeable_backups(self):
      self.setup_directories()
      rounders= [(1, 3), (4, 9)]
      aged_based_data= [
         # rounders:    0   0   0    0    4    4    4  4  4    4    4    4    1   1    1    1    1    1    1 <-- matches era calculation below
         ("file_a", (1000,20  ,  10, 9.1, 8.9, 8.1, 7.9,   6,   5, 4.1, 3.9, 3.1, 2.9,1.9, 1.5, 1.1, 0.9, 0.4, 0.2)),
         ("file_b", (None,19.5)),
         ("file_c", (1001,30  ,None,None,None,None,None,None,None,None,None,None,None,2  )),
         ("file_1d",(1002,31)), # shared checksums
         ("file_2d",(1002,32)),
         ("file_1e",(1003,33)), # shared checksums
         ("file_2e",(1003,None,None,None,None,None,   8)),
      ]
      now= time.time()
      timestamp_lists= {path:[None if age is None else now-age*86400 for age in ages] for path,ages in aged_based_data}
      checksum_lists= {path:[None if t is None else path[-1] +("0"*64 +str(t))[-63:] for t in timestamps]
                       for path,timestamps in timestamp_lists.items()}
      # maps checksums to 3 periods defined by rounders. Period id is maximum number of slots in that period (depending on time of day)
      eras= {}
      for path in timestamp_lists:
         for i, checksum in enumerate(checksum_lists[path]):
            if not checksum: continue
            eras[checksum]= 1 if i<4 else 3 if i<12 else 4
      with database.connect(self.databasepath, self.archivedir) as db:
         runs= [db._make_runname(t+65) for t in timestamp_lists["file_a"]]
         all_checksums= set()
         for path in timestamp_lists:
            for run_name, timestamp, checksum in zip(runs, timestamp_lists[path], checksum_lists[path]):
               if not timestamp: continue
               db.add_backup_entry(run_name, path, timestamp, checksum)
               all_checksums.add(checksum)
         for checksum in all_checksums:
               db.remember_archivedir(checksum, checksum[-2:], True)
         del all_checksums

         result= db.get_purgeable_backups(rounders)
         self.assertTrue(all(a==c[-2:] for c,a,p,t in result)) # correct checksum

         result_checksums= set(c for c,a,p,t in result)

         # the latest versions must never be purged
         latest= set(cl[-1] for cl in checksum_lists.values())
         self.assertEqual(latest & result_checksums, set())

         # file_a
         keep= set(checksum_lists["file_a"]) - result_checksums
         # grouping by era, there must be at most as many versions to keep as slots in era.
         # cannot check "one per slot" directly because slot boundaries depend on time of day
         keep_by_era= {era: set() for era in eras.values()}
         for checksum in keep:
            keep_by_era[eras[checksum]].add(checksum)
         for era in keep_by_era:
            with self.subTest(contents=keep_by_era[era]):
               self.assertLessEqual(len(keep_by_era[era]), era)

         # file_c - one backup in oldest slot is always kept
         purged= [c for c in result_checksums if c.startswith("c")]
         self.assertEqual(purged, [checksum_lists["file_c"][0]]) # 30 and 2 stay

         # file_?d - oldest gets deleted because both paths don't need it any more
         purged= [c for c in result_checksums if c.startswith("d")]
         self.assertEqual(purged, [checksum_lists["file_1d"][0]])
         self.assertEqual(purged, [checksum_lists["file_2d"][0]])

         # file_?e - no deletion because the other still needs it
         purged= [c for c in result_checksums if c.startswith("e")]
         self.assertEqual(purged, [])



# -*- coding: UTF-8 -*-
# Copyright © 2017 Cornelius Grotjahn - This file is covered by the LICENCE file in the project root directory.
import gzip
import os
import sqlite3
import time
import unittest
from abus import config
from abus import crypto
from abus import database
from abus import main
from abus import rebuild
from abus.testbase import AbusTestBase

class RebuildTests(AbusTestBase):
   def _get_database_contents(self):
      """
      reads DB into sets

      :return: set with completed_run, content, and location rows (distinguishable by their lengths)
      """
      queries= ["select run_name, archive_dir, is_complete from run",
                "select run_name, path, timestamp, checksum from content",
                "select checksum, archive_dir, is_compressed from location",
                ]
      result= set()
      with self.get_direct_db_connection() as conn:
         for q in queries:
            result.update(conn.execute(q))
      return result

   def test_rebuild_e2e(self):
      self.setup_multiple_backups()
      before= self._get_database_contents()
      os.unlink(self.databasepath)
      with database.connect(self.databasepath, self.archivedir, allow_create=True):
         pass
      rc= main.main(["test", "-f", self.configfile, "--rebuild-index"])
      self.assertEqual(rc, 0)
      after= self._get_database_contents()
      self.assertEqual(after, before)

   def test_timestamp_string_gets_converted_properly(self):
      # timestamp stores as 1508532089.993785 if you let sqlite3 convert
      self.setup_directories()
      with database.connect(self.databasepath, self.archivedir) as db:
         db.rebuild_content("name", "00/00", [['checksum', '1508532089.9937847', 'path']])
      t= self.executesql("select timestamp from content")
      self.assertEqual(len(t), 1)
      self.assertEqual(t[0][0], 1508532089.9937847)

   records= [
         ("good", "47/11", 1),
         ("good", "/home/somepath", 1507934953.0, "f8e920545e99cdc9bbc2650eb8282344e8971a7ff0c397c91355d0fcaf6c61fa"),
         ("good", "/home/other.path", 1507934453.0, "bcf1a11950d1a1dd1bbe6f3b97cbeb431653e2ad43e1ab42c9ea13f4153764a8"),
         ("incomplete", "47/11", 1),
         ("incomplete", "/home/somepath", 1290000000.0, "32b772fb46ad237546d030bfd880d48220c57c555912f24873debfe8d0505ba2"),
         ("incomplete", "/home/path", 1290003000.0, "498630374d56ea4c53c5baaee00b630be09ca5dbe678b11b1a81b37f1058635f"),
         ("error", "47/11", 0),
         ("error", "/home/somepath", 1507934953.0, "f8e920545e99cdc9bbc2650eb8282344e8971a7ff0c397c91355d0fcaf6c61fa"),
         ("32b772fb46ad237546d030bfd880d48220c57c555912f24873debfe8d0505ba2", "47/11", 1),
         ("f8e920545e99cdc9bbc2650eb8282344e8971a7ff0c397c91355d0fcaf6c61fa", "47/11", 1),
         ("bcf1a11950d1a1dd1bbe6f3b97cbeb431653e2ad43e1ab42c9ea13f4153764a8", "47/11", 0),
      ]
   def run_rebuild(self, records, expected_updates, expected_inserts, expected_deletes):
      """
      Performs a complete rebuild run, filling the database from the parameters first.
      The rebuild should always end up at the state described by self.records.

      :param records: records to for initialising database
      :type records: iterable
      """
      self.setup_directories()
      d= self.archivedir + "/47/11"
      os.makedirs(d)
      with crypto.open_txt(d+"/incomplete.lst", "w", self.password) as f:
         print("32b772fb46ad237546d030bfd880d48220c57c555912f24873debfe8d0505ba2 1290000000.0 /home/somepath", file=f)
         print("498630374d56ea4c53c5baaee00b630be09ca5dbe678b11b1a81b37f1058635f 1290003000.0 /home/path", file=f)
      with crypto.open_txt(d+"/error.lst", "w", self.password) as f:
         print("f8e920545e99cdc9bbc2650eb8282344e8971a7ff0c397c91355d0fcaf6c61fa 1507934953.0 /home/somepath", file=f)
         print("error 0 /home/other.path", file=f)
      with crypto.open_txt(d+"/good.lst", "w", self.password) as f:
         print("f8e920545e99cdc9bbc2650eb8282344e8971a7ff0c397c91355d0fcaf6c61fa 1507934953.0 /home/somepath", file=f)
         print("bcf1a11950d1a1dd1bbe6f3b97cbeb431653e2ad43e1ab42c9ea13f4153764a8 1507934453.0 /home/other.path", file=f)
      with crypto.open_txt(d+"/32b772fb46ad237546d030bfd880d48220c57c555912f24873debfe8d0505ba2.z", "w", self.password) as f:
         f.write("data")
      with crypto.open_txt(d+"/f8e920545e99cdc9bbc2650eb8282344e8971a7ff0c397c91355d0fcaf6c61fa.z", "w", self.password) as f:
         f.write("data")
      with crypto.open_txt(d+"/bcf1a11950d1a1dd1bbe6f3b97cbeb431653e2ad43e1ab42c9ea13f4153764a8", "w", self.password) as f:
         f.write("compressed data")
      with gzip.open(self.archivedir+"/good.gz", "wt") as f:
         print("47/11/32b772fb46ad237546d030bfd880d48220c57c555912f24873debfe8d0505ba2.z", file=f)
         print("47/11/f8e920545e99cdc9bbc2650eb8282344e8971a7ff0c397c91355d0fcaf6c61fa.z", file=f)
         print("47/11/incomplete.lst", file=f)
         print("47/11/error.lst", file=f)
         print("47/11/good.lst", file=f)
         print("good.gz", file=f)
         print("index.sl3", file=f)
      with self.get_direct_db_connection() as conn:
         run_records= set(rac for rac in records if len(rac)==3 and len(rac[0])<60)
         conn.executemany("insert into run(run_name,archive_dir,is_complete) values(?,?,?)", run_records)
         content_records= (t for t in records if len(t)==4)
         conn.executemany("insert into content(run_name, path, timestamp, checksum) values(?,?,?,?)", content_records)
         location_records= (t for t in records if len(t)==3 and len(t[0])==64)
         conn.executemany("insert into location(checksum, archive_dir, is_compressed) values(?,?,?)", location_records)
      cfg= config.Configuration(["test", "-f", self.configfile])
      updates, inserts, deletes = rebuild.rebuild_index_db(cfg)
      expected_db= set(self.records)
      actual= self._get_database_contents()
      self.assertEqual(actual, expected_db)
      self.assertEqual(updates, expected_updates)
      self.assertEqual(inserts, expected_inserts)
      self.assertEqual(deletes, expected_deletes)

   def test_rebuild_no_changes(self):
      r= self.records
      self.run_rebuild(r, 0,0,0)

   def test_rebuild_missing_record(self):
      for i in range(9):
         with self.subTest(i=i):
            r= self.records[:i] + self.records[i+1:]
            self.run_rebuild(r, 0, 1, 0)

   def test_rebuild_changed_record(self):
      changes= [
         (0, "good", "08/15", 1),
         (0, "good", "47/11", 0),
         (1, "good", "/home/somepath", 1111111111.0, "f8e920545e99cdc9bbc2650eb8282344e8971a7ff0c397c91355d0fcaf6c61fa"),
         (1, "good", "/home/somepath", 1507934953.0, "caf956254790e0f71ad80e57b5a69a61183ab2cf9f5b75c406a02d29d6ba3eb5"),
         (2, "good", "/home/other.path", 1111111111.0, "bcf1a11950d1a1dd1bbe6f3b97cbeb431653e2ad43e1ab42c9ea13f4153764a8"),
         (2, "good", "/home/other.path", 1507934453.0, "caf956254790e0f71ad80e57b5a69a61183ab2cf9f5b75c406a02d29d6ba3eb5"),
         (4, "incomplete", "/home/somepath", 1111111111.0, "32b772fb46ad237546d030bfd880d48220c57c555912f24873debfe8d0505ba2"),
         (4, "incomplete", "/home/somepath", 1290000000.0, "caf956254790e0f71ad80e57b5a69a61183ab2cf9f5b75c406a02d29d6ba3eb5"),
         (5, "incomplete", "/home/path", 1111111111.0, "498630374d56ea4c53c5baaee00b630be09ca5dbe678b11b1a81b37f1058635f"),
         (5, "incomplete", "/home/path", 1290003000.0, "caf956254790e0f71ad80e57b5a69a61183ab2cf9f5b75c406a02d29d6ba3eb5"),
         (7, "error", "/home/somepath", 1111111111.0, "f8e920545e99cdc9bbc2650eb8282344e8971a7ff0c397c91355d0fcaf6c61fa"),
         (7, "error", "/home/somepath", 1507934953.0, "caf956254790e0f71ad80e57b5a69a61183ab2cf9f5b75c406a02d29d6ba3eb5"),
         (8, "32b772fb46ad237546d030bfd880d48220c57c555912f24873debfe8d0505ba2", "08/15", 1),
         (8, "32b772fb46ad237546d030bfd880d48220c57c555912f24873debfe8d0505ba2", "47/11", 0),
         ]
      for iteration, change in enumerate(changes):
         with self.subTest(i=iteration):
            r= list(self.records)
            i= change[0]
            r[i]= change[1:]
            self.run_rebuild(r, 1, 0, 0)

   def test_rebuild_extra_location(self):
      r= self.records + [
            ("caf956254790e0f71ad80e57b5a69a61183ab2cf9f5b75c406a02d29d6ba3eb5", "08/15", 1),
         ]
      self.run_rebuild(r, 0,0,1)

   def test_rebuild_extra_entry(self):
      r= self.records + [
            ("good", "/home/completely.different", 1507934453.0, "caf956254790e0f71ad80e57b5a69a61183ab2cf9f5b75c406a02d29d6ba3eb5"),
         ]
      self.run_rebuild(r, 0,0,1)

   def test_rebuild_extra_run(self):
      r= self.records + [
            ("extra", "08/15", 1),
            ("extra", "/home/completely.different", 1507934453.0, "caf956254790e0f71ad80e57b5a69a61183ab2cf9f5b75c406a02d29d6ba3eb5"),
         ]
      self.run_rebuild(r, 0,0,2)

   def test_rebuild_spaces_in_filesname(self):
      self.setup_directories()
      d= self.archivedir + "/47/11"
      os.makedirs(d)
      with crypto.open_txt(d+"/2017_10_31_1232.lst", "w", self.password) as f:
         print("32b772fb46ad237546d030bfd880d48220c57c555912f24873debfe8d0505ba2 1290000000.0 /home/sicko/hello world.txt", file=f)
         print("498630374d56ea4c53c5baaee00b630be09ca5dbe678b11b1a81b37f1058635f 1290003000.0 /home/sicko/real  sick.txt", file=f)
      with gzip.open(self.archivedir+"/2017_10_31_1232.gz", "wt") as f:
         print("47/11/2017_10_31_1232.lst", file=f)
      rc= main.main(["test", "-f", self.configfile, "--rebuild-index"])
      self.assertEqual(rc, 0)
      self.assertEqual(self._get_database_contents(),
                       {("2017_10_31_1232", "47/11", 1),
                        ("2017_10_31_1232", "/home/sicko/hello world.txt", 1290000000.0, "32b772fb46ad237546d030bfd880d48220c57c555912f24873debfe8d0505ba2" ),
                        ("2017_10_31_1232", "/home/sicko/real  sick.txt",  1290003000.0, "498630374d56ea4c53c5baaee00b630be09ca5dbe678b11b1a81b37f1058635f")})




# -*- coding: UTF-8 -*-
# Copyright © 2017 Cornelius Grotjahn - This file is covered by the LICENCE file in the project root directory.

from .misc import set_indentation, cog_make_struct_type, partial_write_guard
from .schapp import schapp_upgrade



# -*- coding: UTF-8 -*-
# Copyright © 2017 Cornelius Grotjahn - This file is covered by the LICENCE file in the project root directory.
import sys
from . import main

def entry_point():
   return main.main(sys.argv)



# -*- coding: UTF-8 -*-
# Copyright © 2017 Cornelius Grotjahn - This file is covered by the LICENCE file in the project root directory.
import logging
import os
import sys
import traceback

from abus import config
from abus import backup
from abus import database
from abus import purge
from abus import rebuild
from abus import restore

def main(argv):
   try:
      cfg= config.Configuration(argv)
      if cfg.is_backup:
         with database.connect(cfg.database_path, cfg.archive_root_path) as db:
            try:
               backup.do_backup(cfg, db)
               purge.do_purge(cfg, db)
            finally:
               purge.write_content_file(cfg, db)
      elif cfg.is_restore:
         restore.restore(cfg)
      elif cfg.is_init:
         if os.path.exists(cfg.database_path):
            raise config.ConfigurationError(cfg.database_path+" exists already")
         os.makedirs(cfg.archive_root_path, exist_ok=True)
         with database.connect(cfg.database_path, cfg.archive_root_path, allow_create=True):
            print("created", cfg.database_path)
      elif cfg.is_rebuild:
         rebuild.rebuild_index_db(cfg)
      else:
         restore.list_archive(cfg)
   except config.ConfigurationError as e:
      error_message= str(e)
      rc= 2
   except RuntimeError as e:
      error_message= str(e)
      rc= 1
   except Exception as e:
      error_message= "".join(traceback.format_exception(type(e), e, e.__traceback__))
      rc= 4
   else:
      logging.shutdown()
      return 0

   print("Error:", error_message, file=sys.stderr)
   if logging.root.handlers: # tests that logging could be initialised before error occurred
      logging.error(error_message)

   logging.shutdown()
   return rc




# -*- coding: UTF-8 -*-
# Copyright © 2017 Cornelius Grotjahn - This file is covered by the LICENCE file in the project root directory.
import gzip
import logging
import os
from   pprint import pprint
import time

from abus.cornelib import partial_write_guard
from abus import config
from abus import database

def do_purge(cfg: config.Configuration, db: database.Database):
   """
   Removes all purgeable backup files.
   """
   logging.info("starting purge")
   purgeable_backups= db.get_purgeable_backups(cfg.retention)
   previous= None
   for checksum, archive_dir, path, timestamp in purgeable_backups:
      time_str= time.strftime("%Y-%m-%d %H:%M", time.localtime(timestamp))
      if checksum==previous:
         logging.info("     == %s from %s", path, time_str)
      else:
         logging.info("Purging %s from %s", path, time_str)
         previous= checksum
         base= cfg.archive_root_path+"/"+archive_dir+"/"+checksum
         if os.path.exists(base+".z"): os.unlink(base+".z")
         if os.path.exists(base): os.unlink(base)
         db.remove_location_entry(checksum)
   purgeable_runs= list(db.get_purgeable_runs())
   for run_name, archive_dir in purgeable_runs:
      index_file_path= cfg.mk_archive_path(run_name, archive_dir, ".lst")
      logging.info("Purging run %s", run_name)
      os.unlink(index_file_path)
   db.purge_runs(r for r,a in purgeable_runs)
   logging.info("completed purge")

def write_content_file(cfg: config.Configuration, db: database.Database) -> None:
   """
   Writes a new content file listing all files in backup directory, removing any previous such file.
   """
   all_runs= list(db.get_all_runs())
   contentfile_name= max(run_name for run_name, subdir in all_runs)+".gz"
   contentfile_path= cfg.archive_root_path+"/"+contentfile_name
   with partial_write_guard(contentfile_path) as partial_path:
      with gzip.open(partial_path, "wt", encoding="UTF=8", newline="\n") as stream:
         try:
            indexdb_rel= os.path.relpath(cfg.database_path, cfg.archive_root_path).replace('\\','/')
         except ValueError:
            pass # database is on a different drive
         else:
            if not indexdb_rel.startswith("../"):
               print(indexdb_rel, file=stream)
         print(contentfile_name, file=stream)
         for run_name, subdir in all_runs:
            print(subdir+"/"+run_name+".lst", file=stream)
         for checksum, subdir, is_compressed in db.get_all_backup_files():
            print(subdir+"/"+checksum+(".z" if is_compressed else ""), file=stream)
   doomed= [
      direntry.path
      for direntry in os.scandir(cfg.archive_root_path)
      if direntry.is_file() and direntry.name.endswith(".gz") and direntry.name != contentfile_name]
   for p in doomed:
      os.unlink(p)
   logging.info("wrote content file "+contentfile_path)



# -*- coding: UTF-8 -*-
# Copyright © 2017 Cornelius Grotjahn - This file is covered by the LICENCE file in the project root directory.
import logging
import os
from abus.testbase import AbusTestBase
from abus import backup
from abus import database
from abus import main

class UpgradeTests(AbusTestBase):
   def check_tables_and_columns(self):
      tables= self.executesql("select name,sql from sqlite_master where type='table'")
      schema= {}
      for (name,sql) in tables:
         joined= sql.replace("\n", " ")
         create,bar,column_defs_string= joined.partition("(")
         column_defs= column_defs_string.split(",")
         columns= set(s.split()[0] for s in column_defs)
         columns.difference_update({"constraint"}) # not a proper column
         schema[name]= columns
      expected_schema= {
         "run": {"run_name", "is_complete", "archive_dir"},
         "content": {"run_name", "path", "timestamp", "checksum"},
         "location": {"checksum", "archive_dir", "is_compressed"},
         "checksum_cache": {"dev", "ino", "timestamp", "checksum"},
         "_database_version": {"current_version"},
      }
      self.assertDictEqual(schema, expected_schema)
      tbl= self.executesql("select * from _database_version")
      self.assertEqual(tbl, [("3.7",)])

   def test_create(self):
      self.setup_directories()
      self.check_tables_and_columns()

   def upgrade_and_check_tables_columns_and_rebuilt_data(self):
      """For upgrades from schema <3, generates a few archive files, whose names must match data
      created by caller, then upgrades and checks that is_compressed and run.archive_dir are
      reconstructed from the filesnames."""
      for f in ["/47/11/2017_10_27_0810.lst",
                "/34/15/2015_06_08_1200.lst",
                "/08/15/32b772fb46ad237546d030bfd880d48220c57c555912f24873debfe8d0505ba2",
                "/13/22/f8e920545e99cdc9bbc2650eb8282344e8971a7ff0c397c91355d0fcaf6c61fa.z",
               ]:
         abspath= self.archivedir+f
         os.makedirs(os.path.dirname(abspath), exist_ok=True)
         with open(abspath, "w"): pass
      with database.connect(self.databasepath, self.archivedir):
         pass
      self.check_tables_and_columns()
      tbl= self.executesql("select run_name, archive_dir, is_complete from run")
      self.assertEqual(set(tbl), {("2017_10_27_0810","47/11",1),("2015_06_08_1200","34/15",0)})
      tbl= self.executesql("select checksum, is_compressed from location")
      self.assertEqual(set(tbl), {("32b772fb46ad237546d030bfd880d48220c57c555912f24873debfe8d0505ba2", 0),
                                  ("f8e920545e99cdc9bbc2650eb8282344e8971a7ff0c397c91355d0fcaf6c61fa", 1),})

   def test_upgrade_v1(self):
      self.setup_directories()
      os.unlink(self.databasepath)
      with open(os.path.dirname(__file__)+"/schema-1.sql") as stream:
         schema_str= stream.read()
      with self.get_direct_db_connection() as conn:
         for sql in schema_str.split(";"):
            conn.execute(sql)
         conn.executemany("insert into run(run_name, timestamp)values(?,?)",
                          [("2017_10_27_0810", 56), ("2015_06_08_1200", None)])
         conn.executemany("insert into location(checksum,archive_dir) values(?,?)",
                          [("32b772fb46ad237546d030bfd880d48220c57c555912f24873debfe8d0505ba2", "08/15"),
                           ("f8e920545e99cdc9bbc2650eb8282344e8971a7ff0c397c91355d0fcaf6c61fa", "13/22")])
      self.upgrade_and_check_tables_columns_and_rebuilt_data()

   def _create_v2_db(self):
      self.setup_directories()
      os.unlink(self.databasepath)
      with self.get_direct_db_connection() as conn:
         conn.execute("""
            create table content(
               run_name text not null,
               path text not null,
               timestamp float not null,
               checksum text not null)""")
         conn.execute("""
            create table location(
               checksum text not null primary key,
               archive_dir text not null)""")
         conn.execute("""
            create table checksum_cache(
               dev int not null,
               ino int not null,
               timestamp float not null,
               checksum text not null)""")
         conn.execute("""
            create table completed_run(
               run_name text not null primary key)""")
         conn.execute("insert into completed_run(run_name) values(?)", ("2017_10_27_0810",))
         conn.executemany("insert into content(run_name, path, timestamp, checksum) values(?,?,?,?)",
                          [("2017_10_27_0810", "c:/some/path", 56, "32b772fb46ad237546d030bfd880d48220c57c555912f24873debfe8d0505ba2"),
                           ("2015_06_08_1200", "C:/bla/bla",   97, "f8e920545e99cdc9bbc2650eb8282344e8971a7ff0c397c91355d0fcaf6c61fa"),
                           ])
         conn.executemany("insert into location(checksum,archive_dir) values(?,?)",
                          [("32b772fb46ad237546d030bfd880d48220c57c555912f24873debfe8d0505ba2", "08/15"),
                           ("f8e920545e99cdc9bbc2650eb8282344e8971a7ff0c397c91355d0fcaf6c61fa", "13/22")])

   def test_upgrade_v2(self):
      self._create_v2_db()
      self.upgrade_and_check_tables_columns_and_rebuilt_data()

   def test_recover_from_upgrade_error(self):
      self._create_v2_db()
      with self.get_direct_db_connection() as conn:
         conn.execute("""alter table location rename to ex_location""")
         conn.execute("""
            create table location(
               checksum text not null,
               archive_dir text not null)""")
         conn.execute("""insert into location select checksum, archive_dir from ex_location""")
         # introducing unique constraint violation after upgrade:
         conn.execute("""insert into location select checksum, archive_dir from ex_location""")

      checksum_before= backup.calculate_checksum(self.databasepath)
      with self.assertLogs() as logs:
         rc= main.main(["test", "-f", self.configfile])
         self.assertEqual(rc, 4)
         self.assertTrue(any(r.levelno==logging.ERROR and "sqlite3.IntegrityError" in r.msg
                             for r in logs.records))
         self.assertTrue(any("index database upgrade failed, restoring backup" in r.msg
                             for r in logs.records))
      checksum_after= backup.calculate_checksum(self.databasepath)
      self.assertEqual(checksum_after, checksum_before)



# -*- coding: UTF-8 -*-
# Copyright © 2017 Cornelius Grotjahn - This file is covered by the LICENCE file in the project root directory.
import concurrent.futures
import enum
import hashlib
import logging
import math
import os
import sys
import threading
import time
import traceback
from   typing import Optional
from contextlib import contextmanager
from fnmatch import fnmatch

from abus.cornelib import partial_write_guard
from abus import config
from abus import crypto
from abus import database
from abus import synchronisation

class Counters(enum.Enum):
   DIRS= "directories"
   UNTOUCHED= "untouched files"
   CHECKSUMMED_ONLY= "already backed up"
   COMPRESSED= "compressed"
   UNCOMPRESSED= "copied"
   ERROR= "errors"

class IndexFile(object):
   def __init__(self, stream):
      self.lock= threading.Lock()
      self.stream= stream
   def add_entry(self, checksum, timestamp, path):
      with self.lock:
         print(checksum, timestamp, path, file=self.stream)

@contextmanager
def open_index_file(run_name, cfg, archive_dir):
   d= cfg.archive_root_path+ "/"+archive_dir
   index_file_path= d+ "/"+run_name+".lst"
   partial_path= index_file_path+".part"
   os.makedirs(d, exist_ok=True)
   try:
      with crypto.open_txt(partial_path, "w", cfg.password) as stream:
         yield IndexFile(stream)
   except:
      os.unlink(partial_path)
      raise
   else:
      os.rename(partial_path, index_file_path)

def read_blocks(stream):
   """returns contents of binary files as list of byte-blocks"""
   block_size= 8192
   while True:
      block= stream.read(block_size)
      yield block
      if len(block) < block_size:
         break

def do_backup(cfg: config.Configuration, db: database.Database) -> None:
   """Performs a backup run with given configuration.
   """
   initial_subdir_usage= db.get_archivedir_usage()
   iter_archivedirs_to_use= iter(archive_dirs_to_use(initial_subdir_usage))
   run_file_archive_dir= next(iter_archivedirs_to_use)
   run_name= db.open_backup_run(run_file_archive_dir)
   logging.info("starting backup run %s", run_name)
   stats= {c:0 for c in Counters}
   n_threads= 20
   with concurrent.futures.ThreadPoolExecutor(n_threads) as executor, \
         open_index_file(run_name, cfg, run_file_archive_dir) as index_file:
      # This thread walks the directories, runs do_one_file in the background for each file
      # and then waits for all the background tasks.
      futures= []
      dir_queue= list(cfg.include)
      while dir_queue:
         p= dir_queue.pop()
         try:
            direntries= os.scandir(p)
         except (PermissionError, OSError) as e:
            msg= str(e)
            displaydir= "" if p in msg else " "+p
            logging.error("Error reading directory%s: %s", displaydir, msg)
            stats[Counters.ERROR] += 1
            continue
         else:
            stats[Counters.DIRS] += 1
         for direntry in direntries:
            path= direntry.path.replace('\\', '/')
            if any(fnmatch(path, p) for p in cfg.exclude):
               pass
            elif direntry.is_file(follow_symlinks=False):
               # restricts number of simultaneous single-file backups:
               if len(futures)>=n_threads*2:
                  wait_for_one(futures, stats)
               f= executor.submit(do_one_file, direntry, run_name, iter_archivedirs_to_use, cfg, db, index_file)
               futures.append(f)
            elif direntry.is_dir(follow_symlinks=False):
               dir_queue.append(path)
      while futures:
         wait_for_one(futures, stats)
   logging.info(", ".join("{}: {}".format(c.value, n) for c,n in stats.items()))
   n_errors= stats[Counters.ERROR]
   if n_errors==0:
      db.complete_backup_run(run_name)
      logging.info("completed backup run %s without errors", run_name)
   else:
      logging.warning("completed backup run %s with errors", run_name)
      raise RuntimeError("some files failed to be backed up")

def wait_for_one(futures, stats):
   done,not_done = concurrent.futures.wait(futures, return_when=concurrent.futures.FIRST_COMPLETED)
   for f in done:
      stats[f.result()] += 1
   futures[:]= not_done

def archive_dirs_to_use(initial_subdir_usage):
   """
   Returns sequence of archive subdirectories (relative to root) to be used for each single new file in turn.
   The dirs returned may or may not exist.

   :param initial_subdir_usage: dict existing archive_dir -> n files therein
   """
   files_per_dir= 100
   for archive_dir, n in initial_subdir_usage.items():
      for _ in range(n, files_per_dir):
         yield archive_dir
   def infrange():
      n= 0
      while True:
         yield n
         n += 1
   def mkdirname(n):
      least= "{:02}".format(n%100)
      return least if n<100 else least + "/" + mkdirname(n//100)
   for n in infrange():
      archive_dir= mkdirname(n)
      if archive_dir not in initial_subdir_usage:
         for _ in range(files_per_dir):
            yield archive_dir

def calculate_checksum(path: str) -> str:
   """Calculates SHA256 sum of a file"""
   partially_calculated_checksum= hashlib.sha256()
   with open(path, "rb") as fd:
      for blk in read_blocks(fd):
         partially_calculated_checksum.update(blk)
   return partially_calculated_checksum.hexdigest()

class FileChangedWhileReadingError(Exception):
   pass

class Cache:
   def __init__(self, fun):
      self.f= fun
      self.v= None
   def __call__(self):
      if self.v is None:
         self.v= self.f()
      return self.v

def do_one_file(direntry, run_name, iter_archivedirs_to_use, cfg, db, index_file):
   """
   Processes a single file, making appropriate database entries for it and creating the
   backup copy if necessary.

   :param direntry: a direntry with the file information as returned by os.scandir()
   :param run_name: the backup run to which the file should be added
   :param iter_archivedirs_to_use: iterator over archive dirs (relative to archive root)
                                   into which new backup copies should be placed.
   :type cfg: config.Configuration
   :type db: database.Database
   :param index_file: Open file descriptor for writing the index entry
   :type index_file: IndexFile
   """
   path= direntry.path.replace("\\", "/")
   try:
      stat= direntry.stat()
      if stat.st_dev==0 and stat.st_ino==0:
         # this means this is windows and the direntry does not have the relevant information
         stat= os.stat(path)

      archive_dir_fun= Cache(lambda:next(iter_archivedirs_to_use))
      failed_attempts= 0
      while True: # retry loop
         try:
            result= do_one_file_no_retry(run_name, path, stat, cfg, db, archive_dir_fun, index_file)
         except FileChangedWhileReadingError:
            failed_attempts += 1
            if failed_attempts==5:
               raise RuntimeError("file changed while reading")
            time.sleep(failed_attempts)
            stat= os.stat(path)
         else:
            return result
   except Exception as e:
      index_file.add_entry("error", 0, path)
      log_error(e, direntry)
      return Counters.ERROR

def do_one_file_no_retry(run_name, path, stat, cfg, db, archive_dir_fun, index_file):
   result= Counters.UNTOUCHED
   with synchronisation.global_data_lock(stat.st_dev, stat.st_ino, stat.st_mtime):
      checksum= db.lookup_checksum(stat.st_dev, stat.st_ino, stat.st_mtime)
      if checksum is None:
         checksum= calculate_checksum(path)
         db.remember_checksum(stat.st_dev, stat.st_ino, stat.st_mtime, checksum)
         result= Counters.CHECKSUMMED_ONLY

   with synchronisation.global_data_lock(checksum):
      if not db.have_checksum(checksum):
         backup_path= cfg.backup_path_abs(archive_dir_fun(), checksum)
         if stat.st_size < cfg.minimum_size_for_compression or os.path.splitext(path)[-1] in cfg.compressed_extensions:
            open_dst_function= crypto.open_bin
            result= Counters.UNCOMPRESSED
            is_compressed= False
         else:
            backup_path += ".z"
            open_dst_function= crypto.open_lzma
            result= Counters.COMPRESSED
            is_compressed= True
         make_backup_copy(path, checksum, backup_path, open_dst_function, cfg.password)
         db.remember_archivedir(checksum, archive_dir_fun(), is_compressed)
   db.add_backup_entry(run_name, path, stat.st_mtime, checksum)
   index_file.add_entry(checksum, stat.st_mtime, path)
   return result

def log_error(e: Exception, direntry: os.DirEntry) -> None:
   """
   Issues as pretty an error message as possible

   :param direntry: of file that we tried to back up when exception occurred
   """
   #logging.debug("%s", "".join(traceback.format_exception(type(e), e, e.__traceback__)))
   if isinstance(e, RuntimeError):
      error_msg= str(e)
   else:
      error_msg= "".join(traceback.format_exception(type(e), e, e.__traceback__))
   path= direntry.path.replace("\\","/")
   if path.lower() not in error_msg.replace("\\","/").lower():
      error_msg= "While backing up {}: {}".format(path, error_msg)
   logging.error(error_msg)
   print(error_msg, file=sys.stderr)

def make_backup_copy(path, expected_checksum, backup_path, open_dst_function, password):
   """
   Creates the encrypted backup copy of a file, taking care not to leave a partial backup file in case of errors.

   :param path: Path of file to be backed up.
   :param expected_checksum: Known checksum of the file, which is verified while copying; if different an exception
                             will be raised.
   :param backup_path: absolute filename of the destination file
   :param open_dst_function: function from crypto module to open the destination file, i.e. either open_lzma or open_bin according to whether
                             compression is required.
   :param password: to be used for encryption
   """
   os.makedirs(os.path.dirname(backup_path), exist_ok=True)
   partially_calculated_checksum= hashlib.sha256()
   with partial_write_guard(backup_path) as partial_path:
      with open(path,"rb") as src, open_dst_function(partial_path, "w", password) as dst:
         for blk in read_blocks(src):
            partially_calculated_checksum.update(blk)
            dst.write(blk)
      if partially_calculated_checksum.hexdigest() != expected_checksum:
         raise FileChangedWhileReadingError()



# -*- coding: UTF-8 -*-
# Copyright © 2017 Cornelius Grotjahn - This file is covered by the LICENCE file in the project root directory.
import gzip
import logging
import os
import re
from pprint import pprint
from typing import Iterable,Optional,Set,Tuple

from abus import config
from abus import crypto
from abus import database

def rebuild_index_db(cfg):
   """
   Reconstructs the location table from scanning the archive dir

   :type cfg: config.Configuration
   """
   logging.info("rebuilding index database")
   with database.connect(cfg.database_path, cfg.archive_root_path) as db:
      actual_backup_files= set(_find_archive_files(cfg.archive_root_path))
      _validate_against_content_file(cfg, actual_backup_files)
      location_data= _filter_backup_files(actual_backup_files)
      updates, inserts, deletes = db.rebuild_location_table(location_data)

      runs= list(_filter_run_files(actual_backup_files))
      for run_name, archive_dir in runs:
         index_file_path= cfg.mk_archive_path(run_name, archive_dir, ".lst")
         with crypto.open_txt(index_file_path, "r", cfg.password) as index_file:
            splut_lines= (line.strip().split(maxsplit=2) for line in index_file)
            u,i,d = db.rebuild_content(run_name, archive_dir, splut_lines)
         updates += u; inserts += i; deletes += d
      deletes += db.remove_runs(other_than=(r[0] for r in runs))
   logging.info("index rebuild added %d, removed %d, changed %d entries", inserts, deletes, updates)
   return updates, inserts, deletes

def _validate_against_content_file(cfg: config.Configuration, actual_backup_files: Set[Tuple[str,str]]):
   """
   Checks that all path from latest content file are actually in backup archive.

   :param actual_backup_files: Paths (filename, archive_dir) that are actually in the backup archive."""
   content_files= sorted(filename for filename,archive_dir in actual_backup_files
                         if filename.endswith(".gz") and archive_dir=="")
   if len(content_files)<1:
      raise RuntimeError("Could not find a content file.")
   with gzip.open(cfg.mk_archive_path(content_files[-1]), "rt") as stream:
      for line in stream:
         archivedir, sep, filename = line[:-1].rpartition('/')
         if (filename,archivedir) not in actual_backup_files:
            raise RuntimeError("Path {} missing from backup archive".format(line[:-1]))

def _filter_run_files(archive_contents: Iterable[Tuple[str, str]]) -> Iterable[Tuple[str, str]]:
   """
   Filters run-files from all paths in backup archive.

   :param archive_contents: all (filename, archive_dir)s in backup archive.
   :return: (run_name, archive_dir)s
   """
   for filename, archive_dir in archive_contents:
      run_name, ext = os.path.splitext(filename)
      if ext==".lst":
         yield run_name, archive_dir

def find_index_files(archive_root_path: str) -> Iterable[Tuple[str,str]]:
   """
   returns iterable of .lst files

   :param archive_root_path: from cfg
   :type db: database.Database
   :return: (run_name, archive_dir)s
   """
   return _filter_run_files(_find_archive_files(archive_root_path))

def _find_archive_files(archive_root_path: str) -> Iterable[Tuple[str, str]]:
   """
   Returns list of files in archive dir.

   :param archive_root_path: from cfg
   :return: (filename, archive_dir)s
   """
   dirqueue= []
   dirpath, archive_dir = archive_root_path, ""
   while True:
      for direntry in os.scandir(dirpath):
         if direntry.is_dir():
            sepd= "" if archive_dir=="" else archive_dir+"/"
            dirqueue.append((direntry.path, sepd + direntry.name))
         elif direntry.is_file():
            yield direntry.name, archive_dir
      if not dirqueue: break
      dirpath, archive_dir = dirqueue.pop()

def _filter_backup_files(archive_contents: Iterable[Tuple[str, str]]) -> Iterable[Tuple[str, str, bool]]:
   """
   Filters backup files from all paths in backup archive.

   :param archive_contents: all (filename, archive_dir)s in backup archive.
   :return: (checksum, archive_dir, is_compressed)s
   """
   re_archive_filename= re.compile(r"[0-9a-f]{64}")
   for filename, archive_dir in archive_contents:
      m= re_archive_filename.match(filename)
      if m:
         checksum= m.group(0)
         yield checksum, archive_dir, filename.endswith(".z")

def find_compressed_backups(archive_root_path) -> Iterable[str]:
   """
   Returns list of backup files that are compressed.

   :param archive_root_path: from cfg
   :return: checksums
   """
   for checksum,archive_dir,is_compressed in _filter_backup_files(_find_archive_files(archive_root_path)):
      if is_compressed:
         yield checksum



# -*- coding: UTF-8 -*-
# Copyright © 2017 Cornelius Grotjahn - This file is covered by the LICENCE file in the project root directory.
import os
import time
from abus import config
from abus import main
from abus.testbase import AbusTestBase

class ConfigTestser(AbusTestBase):
   def test_parse_date(self):
      def str2sec(s):
         return time.mktime(time.strptime(s, "%Y%m%d%H%M%S"))
      now= str2sec("20170912150000")
      for d in range(1,9):
         for t in range(8):
            for future_date in True, False:
               for future_time in True, False:
                  d_str= "20171003" if future_date else "20170912"
                  t_str= "-160723" if future_time else "-130905"
                  test_str= d_str[-d:] + t_str[:t]

                  if d in (4,6,8) and t in (0,5,7):
                     expect_str= "2016"+d_str[-4:] if future_date and d==4 else d_str
                     expect_str += (t_str[1:t]+"000000")[:6]
                     try:
                        self.assertEqual(config._parse_date(test_str, now), str2sec(expect_str))
                     except:
                        print("test string:", test_str)
                        raise
                  else:
                     with self.assertRaises(ValueError):
                        print("test string:", test_str, "result:", config._parse_date(test_str, now))

   def test_conflicting_cmdline_actions(self):
      self.setup_directories()
      with self.assertRaisesRegex(config.ConfigurationError, "conflicting command line actions"):
         config.Configuration(["test", "-f", self.configfile, "--backup", "--restore"])

   def test_init_creates_directory(self):
      self.setup_directories()
      os.unlink(self.databasepath)
      os.rmdir(self.archivedir)
      rc= main.main(["test", "-f", self.configfile, "--init"])
      self.assertEqual(rc, 0)
      self.assertTrue(os.path.isfile(self.databasepath))
      # non-standard database file:
      db_path= self.otherdir+"/index_database"
      self.write_config_file("indexdb " +db_path)
      os.unlink(self.databasepath)
      os.rmdir(self.archivedir)
      rc= main.main(["test", "-f", self.configfile, "--init"])
      self.assertEqual(rc, 0)
      self.assertTrue(os.path.isfile(db_path))
      self.assertFalse(os.path.exists(self.databasepath))

   def test_retention_parsing(self):
      self.setup_directories()
      def get_cfg_for_retention(*retention_strings):
         self.write_config_file(*retention_strings)
         return config.Configuration(["test", "-f", self.configfile])

      # default
      cfg= get_cfg_for_retention()
      self.assertEqual(cfg.retention, [(1,7), (56,150)])
      # std
      cfg= get_cfg_for_retention("retain 7 20  1 10  .25 2  28 182")
      self.assertEqual(cfg.retention, [(.25,2), (1,10), (7,20), (28,182)])
      # non-numbers
      with self.assertRaises(config.ConfigurationError):
         get_cfg_for_retention("retain 7 28  56 b")
      # missing arguments
      with self.assertRaises(config.ConfigurationError):
         get_cfg_for_retention("retain 7 28  14 56 28")
      with self.assertRaises(config.ConfigurationError):
         get_cfg_for_retention("retain")
      # duplicates
      with self.assertRaises(config.ConfigurationError):
         get_cfg_for_retention("retain 1 10  7 10")
      # non-factors
      with self.assertRaises(config.ConfigurationError):
         get_cfg_for_retention("retain 4 10  7 20")



# -*- coding: UTF-8 -*-
# Copyright © 2017 Cornelius Grotjahn - This file is covered by the LICENCE file in the project root directory.
import itertools
import os
import sqlite3
from typing import Iterable, Tuple
from abus import backup
from abus import crypto
from abus import database
from abus import main
from abus.testbase import AbusTestBase

class RestoreTests(AbusTestBase):
   def test_database_get_archive(self):
      """Writes various rundata to DB and checks that Database returns the right listings.
      No actual restoring or backing up takes place"""
      self.setup_directories()
      checksums= ("e{:03}".format(i) for i in itertools.count()).__iter__()
      timestamps= itertools.count(1506159900, 50).__iter__()
      all_files= ["c:/home/file_{}.txt".format(c) for c in "abcdefghxy"]
      file_states= {}
      data= [] # (run_name, timestamp, path, checksum, archive_dir) list
      for run,files in [("2017-09-22 12:45","abcdefghxy"),
                        ("2017-09-23 12:45","xy"),
                        ("2017-09-24 12:45","bcdefghx"),
                        ("2017-09-25 12:45","abcfghy"),
                       ]:
         for c in files:
            file_states[c]= timestamps.__next__(), checksums.__next__()
         # this sometimes "touches" b, tests that some content is not listed twice:
         file_states["b"]= timestamps.__next__(), file_states["b"][1]
         data.extend((run,ts,"c:/home/file_{}.txt".format(c),cs,cs[:2]+"/"+cs[2:])
                     for c,(ts,cs) in file_states.items())
      with sqlite3.connect(self.databasepath) as conn:
         conn.executemany("insert into location(checksum, archive_dir, is_compressed) values(?,?,1)", set(t[3:5] for t in data))
         conn.executemany("insert into content(run_name, timestamp, path, checksum) values(?,?,?,?)", (t[0:4] for t in data))
         conn.execute("insert into run(run_name,archive_dir,is_complete) select distinct run_name, '', 1 from content")

      with database.connect(self.databasepath, self.archivedir) as db:
         def test_one(cutoff_date,
                      show_all,
                      nearly_expect: Iterable[Tuple[str,float,str,str,int]]):
            # nearly_expect may still contain duplicate (file,content) with different timestamps
            pac_set= set((p,a,c)for p,t,a,c,z in nearly_expect)
            ptacz_list= [min(ptacz for ptacz in nearly_expect if ptacz[0]==p and ptacz[2]==a and ptacz[3]==c)
                         for p,a,c in pac_set]
            expect= sorted(ptacz_list, key= lambda ptacz: (ptacz[0],-ptacz[1]))
            result= list(db.get_archive_contents([], cutoff_date, show_all))
            self.assertEqual(result, expect)

            expect= [ptac for ptac in expect if ptac[0][-5] in "ab"]
            result= list(db.get_archive_contents(['*a.txt','*b.txt'], cutoff_date, show_all))
            self.assertEqual(result, expect)

            expect= [ptac for ptac in expect if ptac[0][-5]=="a"]
            result= list(db.get_archive_contents(['*a.txt'], cutoff_date, show_all))
            self.assertEqual(result, expect)

         test_one(cutoff_date=None, show_all=False,
                  nearly_expect= [(p,t,a,c,1) for r,t,p,c,a in data if r=="2017-09-25 12:45"],
                  )
         test_one(cutoff_date=None, show_all=True,
                  nearly_expect= [(p,t,a,c,1) for r,t,p,c,a in data],
                  )

         by_path= {}
         for cutoff in sorted(set(t for r,t,p,c,a in data)):
            with self.subTest(cutoff=cutoff):
               test_one(cutoff_date=cutoff, show_all=True,
                        nearly_expect= [(p,t,a,c,1) for r,t,p,c,a in data if t<=cutoff],
                        )
               by_path.update({p:(p,t,a,c,1) for r,t,p,c,a in data if t==cutoff})
               test_one(cutoff_date=cutoff, show_all=False,
                        nearly_expect= by_path.values(),
                        )

   def test_simple_restore(self):
      self.setup_backup_with_well_known_checksums()
      self.remove_dir_contents(self.restoredir)
      rc= main.main(["test", "-f", self.configfile, "-r"])
      self.assertEqual(rc, 0)
      self.assertEqual(sum(1 for _ in self.find_files(self.restoredir)), len(self.expected_backups))
      for relpath, archive_name in self.expected_backups:
         actual= backup.calculate_checksum(self.restoredir+"/"+relpath)
         expected= archive_name[:64]
         self.assertEqual(actual, expected)

   def test_restore_does_not_overwrite(self):
      self.setup_backup_with_well_known_checksums()
      self.remove_dir_contents(self.restoredir)
      filenames= [path.partition("/")[0] for path,_ in self.expected_backups]
      for filename in filenames:
         with open(filename,"w",newline='\n') as f: f.write("Hello World\n")
      rc= main.main(["test", "-f", self.configfile, "-r"])
      self.assertEqual(rc, 0)
      for filename in filenames:
         actual= backup.calculate_checksum(filename)
         self.assertEqual(actual, "d2a84f4b8b650937ec8f73cd8be2c74add5a911ba64df27458ed8229da804a26")

   def test_no_partial_backupfile_is_left(self):
      self.setup_backup_with_well_known_checksums()
      relpath, archivename= max(self.expected_backups)
      path= self.homedir+"/"+relpath
      expected_checksum= "42" # causes Exception which is expected not to leave file behind
      password= "Would you like to buy some air?"
      open_dst_function= crypto.open_lzma
      backup_path= self.archivedir+"/42"
      with self.assertRaises(backup.FileChangedWhileReadingError):
         backup.make_backup_copy(path, expected_checksum, backup_path, open_dst_function, password)
      self.assertFalse(os.path.exists(backup_path))
      self.assertFalse(os.path.exists(backup_path+".part"))
      # making doubly sure the file would normally have been created (the source path might just be wrong, for example):
      backup.make_backup_copy(path, archivename[:64], backup_path, open_dst_function, password)
      self.assertTrue(os.path.exists(backup_path))
      self.assertFalse(os.path.exists(backup_path+".part"))

   def test_allversions_restore(self):
      self.setup_multiple_backups()
      n_versions= sum(1 for direntry in self.find_files(self.archivedir)
                        if len(direntry.name) in(64,66))
      rc= main.main(["test", "-f", self.configfile, "-r", "-a"])
      self.assertEqual(rc, 0)
      self.find_files(self.restoredir)
      n_restored= sum(1 for _ in self.find_files(self.restoredir))
      self.assertEqual(n_restored, n_versions)

   def test_bug_index_error_if_no_files_match(self):
      self.setup_backup_with_well_known_checksums()
      self.remove_dir_contents(self.restoredir)
      rc= main.main(["test", "-f", self.configfile, "-r", self.homedir])
      self.assertEqual(rc, 0)
      self.assertEqual(list(self.find_files(self.restoredir)), [])

   def test_paths_from_different_drive(self):
      self.setup_backup_with_well_known_checksums()
      with sqlite3.connect(self.databasepath) as conn:
         conn.execute("""update content set path= replace(path, 'C:', 'E:') where path like '%/my_%'""")
      rc= main.main(["test", "-f", self.configfile, "-r"])
      self.assertEqual(rc, 0)
      files= list(self.find_files(self.restoredir))
      self.assertEqual(len(files), 4)




# -*- coding: UTF-8 -*-
# Copyright © 2017 Cornelius Grotjahn - This file is covered by the LICENCE file in the project root directory.
from setuptools import setup

with open("README") as f:
   readme= f.read()

setup(name= 'abus',
   version= '7',
   description= 'Abingdon Backup Script',
   long_description= readme,
   author= 'Cornelius Grotjahn',
   author_email= 's1@tempaddr.uk',
   license= 'MIT',
   packages= ['abus','abus/cornelib'],
   package_data={'abus': ["schema-3.sql", "upgrade-2.sql", "upgrade-1.sql"]},
   install_requires= ['cryptography'],
   zip_safe= False,
   entry_points = { 'console_scripts': ['abus=abus:entry_point'], }, # script=package:function
   classifiers= [
      'Programming Language :: Python :: 3.6',
      'Programming Language :: Python :: 3 :: Only',
      'Development Status :: 4 - Beta',
      'License :: OSI Approved :: MIT License',
      'Topic :: System :: Archiving :: Backup',
      ],
   url="https://pypi.python.org/pypi/abus",
   )



# -*- coding: UTF-8 -*-
# Copyright © 2017 Cornelius Grotjahn - This file is covered by the LICENCE file in the project root directory.
import argparse
import logging
import os
import time
from   typing import Optional

class ConfigurationError(Exception):
   pass

class Configuration(object):
   """Combined backup configuration options from command line, environment, and config file"""
   def __init__(self, argv):
      """
      Combined backup configuration options from command line, environment, and config file (determined
      by command line or environment.

      :param argv: command line arguments (`sys.argv`).
      """

      # Default values
      self.archive_root_path= None
      self.database_path= None
      self.logfile_path= None
      self.include= []
      self.exclude= []
      self.password= None
      self.minimum_size_for_compression= 4096
      self.compressed_extensions= set([".jpg", ".tif", ".gz", ".tgz",])
      self.retention= [(1,7), (56,150)]

      # Command line
      args= self._get_command_line_options(argv)
      if args.f is not None:
         config_path= args.f
      elif "ABUS_CONFIG" in os.environ:
         config_path= os.environ["ABUS_CONFIG"]
      else:
         raise ConfigurationError("Missing config file path (-f option or ABUS_CONFIG environment variable)")
      is_list= args.list
      self.is_backup= args.backup
      self.is_restore= args.restore
      self.is_init= args.init
      self.is_rebuild= args.rebuild_index
      self.list_all= args.a
      self.patterns= args.glob
      self.cutoff= args.d

      # Config file
      with open(config_path, encoding="utf-8") as f:
         try:
            self._parse_config_file(f)
         finally:
            # Initialising log file now because it might be possible despite a subsequent
            # error during parsing.
            if self.logfile_path:
               logging.basicConfig(filename=self.logfile_path,
                                   level=logging.DEBUG,
                                   format='%(asctime)s %(levelname)-7s %(message)s',
                                   datefmt='%Y-%m-%d %H:%M:%S',
                                   )

      actions= is_list, self.is_backup, self.is_restore, self.is_init, self.is_rebuild
      n_actions= sum(1 for a in actions if a)
      if n_actions>1:
         raise ConfigurationError("conflicting command line actions")
      if n_actions==0:
         is_list= True
      if (self.list_all or self.cutoff) and not (self.is_restore or is_list):
         raise ConfigurationError("conflicting command line options (-a/-d without -r/-l)")
      if not self.archive_root_path:
         raise ConfigurationError("archive directory option not set")
      if not os.path.isdir(self.archive_root_path) and not self.is_init:
         raise ConfigurationError("missing archive directory ({})".format(self.archive_root_path))
      if not self.password:
         raise ConfigurationError("password directory option not set")
      if not self.include:
         raise ConfigurationError("missing or empty [include] section")
      if not self.logfile_path:
         raise ConfigurationError("logfile path option not set")

      if not self.database_path:
         self.database_path= self.archive_root_path+"/index.sl3"

   def _parse_config_file(self, lines):
      self.include= []
      self.exclude= []
      section= None
      for line in lines:
         line= line.strip()
         if not line or line.startswith('#'):
            pass
         elif line.lower().startswith("[incl"):
            section= self.include
         elif line.lower().startswith("[excl"):
            section= self.exclude
         elif line.lower().startswith("["):
            raise ConfigurationError("unknown section name: "+line)
         elif section is not None:
            line= line.replace("\\", "/")
            if line.endswith("/"): line= line[:-1]
            section.append(line)
         else:
            splut= line.split(maxsplit=1)
            if len(splut)!=2:
               raise ConfigurationError("missing value: "+line)
            keyword, args = splut
            if keyword=="archive":
               self.archive_root_path= args
            elif keyword=="indexdb":
               self.database_path= args
            elif keyword=="logfile":
               self.logfile_path= args
            elif keyword=="password":
               self.password= args
            elif keyword=="retain":
               self.retention= _parse_retention_args(args)
            else:
               raise ConfigurationError("unknown option: "+keyword)

   def _get_command_line_options(self, argv):
      parser= argparse.ArgumentParser(
         prog=os.path.basename(argv[0]),
         description="""
            The Abingdon BackUp Script makes copies of files to a backup location on a local filesystem.
            For documentation, read the accompanying README or the project front page on PyPi.
            """,
         )
      parser.add_argument('-f',
         action='store',
         metavar='path',
         help="""Path to config file required for defining backup location etc.
                 Defaults to value of the ABUS_CONFIG environment variable.""")
      parser.add_argument('--list', '-l',
         action='store_true',
         help='Lists the contens of the backup directory. This is the default action.')
      parser.add_argument('--backup',
         action='store_true',
         help="Backs up files to the backup directory.")
      parser.add_argument('--restore', '-r',
         action='store_true',
         help="Restores files from the backup directory to the current directory.")
      parser.add_argument('-a',
         action='store_true',
         help="""Includes all matching file versions when listing or restoring
            rather than only the latest matching version of each file.""")
      parser.add_argument('-d',
                          action='store',
                          type=_parse_date,
                          metavar="datetime",
                          help="""
            Cut-off time (format [[cc]yy]mmdd[-HHMM[SS]]) from which files are not included in listing or restore.
            The year defaults to the current year or the previous year if date (without time) would be in the future.
            The time defaults to midnight.
            Note that to show all files in May, say, the argument must be 0601.
            """)
      parser.add_argument('--init',
         action='store_true',
         help='Creates an empty index database and the backup directory.')
      parser.add_argument('--rebuild-index',
         action='store_true',
         help="""Reconstructs the index database from static backup files. N.B.: You must convince yourself of the
            integrity of the content file <date>.gz first.""")
      parser.add_argument('glob',
         nargs='*',
         help="""File path pattern (* matches /) of file to be included in listing or restore.
            '/' must be used as a path separator.""")
      return parser.parse_args(argv[1:])


   def backup_path_abs(self, archive_dir_rel, checksum):
      return self.archive_root_path +"/" +archive_dir_rel +"/" +checksum

   def mk_archive_path(self, relpath:str, archive_dir: Optional[str]= None, ext: Optional[str]= None) -> str:
      """
      Makes absolute path from relpath relative to archive root.

      :param archive_dir: will be prepended to path if given
      :param ext: will be appended to relpath if given
      """
      return (self.archive_root_path
              +("/"+archive_dir if archive_dir else "")
              +"/" +relpath
              +(ext if ext else "")
              )

def _parse_date(string, now=None):
   today_str= time.strftime("%Y%m%d", time.localtime(now))
   d_str, minus, t_str = string.partition("-")
   if (t_str=="") == (minus==""):
      if len(d_str) in(4,6,8) and len(t_str) in(0,4,6):
         if d_str.isdigit():
            if t_str=="" or t_str.isdigit():
               full_t_str= t_str + "000000"[len(t_str):]
               full_d_str= today_str[:8-len(d_str)] + d_str
               if len(d_str)==4 and full_d_str > today_str:
                  last_year= int(today_str[:4])-1
                  full_d_str= str(last_year) + d_str
               tm= time.strptime(full_d_str + full_t_str, "%Y%m%d%H%M%S")
               return time.mktime(tm)
   raise ValueError()

def _parse_retention_args(arg_string):
   try:
      values= tuple(float(a) for a in arg_string.split())
   except ValueError:
      raise ConfigurationError("expected numeric arguments: "+arg_string)
   n= len(values)
   if n==0 or n%2==1:
      raise ConfigurationError("expected pairs of numbers: "+arg_string)
   paired= ((values[i],values[i+1]) for i in range(0,n,2))
   retention= sorted(paired, key=lambda ra:ra[1])
   if len(set(age for _,age in retention)) < len(retention):
      raise ConfigurationError("duplicate retention age")
   rounders= [r for r,a in retention]
   for a,b in zip(rounders, rounders[1:]):
      if b % a != 0:
         raise ConfigurationError("retention frequencies {} and {} are not multiples".format(a,b))
   return retention



# -*- coding: UTF-8 -*-
# Copyright © 2017 Cornelius Grotjahn - This file is covered by the LICENCE file in the project root directory.
import concurrent.futures
import os
import time
import traceback

from abus.backup import read_blocks
from abus import crypto
from abus import database
from abus.synchronisation import global_data_lock

def list_archive(config):
   with database.connect(config.database_path, config.archive_root_path) as db:
      results= db.get_archive_contents(patterns=config.patterns,
                                       cutoff_date=config.cutoff,
                                       show_all=config.list_all)
      for path, timestamp, archive_dir, checksum, _is_compressed in results:
         time_str= time.strftime("%Y-%m-%d %H:%M", time.localtime(timestamp))
         print("{}  {}".format(time_str, path))

def restorepath(path: str) -> str:
   """Converts an original path to a restore path by replacing any DOS drive letters with
   a sub directory"""
   if len(path)>=2 and path[1]==':':
      return "/"+path[0].lower()+path[2:]
   else:
      return path

def restore(config):
   with database.connect(config.database_path, config.archive_root_path) as db:
      results= db.get_archive_contents(patterns=config.patterns,
                                       cutoff_date=config.cutoff,
                                       show_all=config.list_all)
      results= ((restorepath(p),t,a,c,z) for p,t,a,c,z in results)
      results= list(results) # path, timestamp, archive_dir, checksum, is_compressed
      if len(results)==0:
         return # have to check because empty generator breaks commonpath()
      common= os.path.commonpath(ptacz[0] for ptacz in results)
      n_threads= 4
      futures= []
      progress= Progress(len(results))
      with concurrent.futures.ThreadPoolExecutor(n_threads) as executor:
         for ptacz in results:
            if len(futures)>=n_threads*2:
               progress.wait_for_one(futures)
            f= executor.submit(restore_one, ptacz, config, common)
            futures.append(f)
         while futures:
            progress.wait_for_one(futures)
   progress.give_summary()

class CannotOverwriteError(RuntimeError):
   pass

def restore_one(ptacz, config, common):
   path, timestamp, archive_dir, checksum, is_compressed = ptacz
   target= os.path.relpath(path, start=common)
   if config.list_all:
      sansext, ext= target= os.path.splitext(target)
      target= sansext +time.strftime("-%Y%m%d-%H%M", time.localtime(timestamp)) +ext
   with global_data_lock(target):
      if os.path.exists(target):
         raise CannotOverwriteError("Cannot overwrite " + target)

      parent= os.path.dirname(target)
      with global_data_lock(parent):
         if parent=="" or os.path.isdir(parent):
            pass # good
         elif os.path.exists(parent):
            raise CannotOverwriteError("Cannot overwrite {} for {}".format(parent, target))
         else:
            os.makedirs(parent)
      source= config.archive_root_path +"/" +archive_dir +"/" +checksum
      if is_compressed:
         src_cm= crypto.open_lzma(source+".z", "r", config.password)
      else:
         src_cm= crypto.open_bin(source, "r", config.password)
      with src_cm as src, open(target, "wb") as dstfd:
         for blk in read_blocks(src):
            dstfd.write(blk)
   return path

class Progress:
   def __init__(self, n_files):
      self.n_files= n_files
      self.t0= time.time()
      self.times= []   # (t,n)s where n files have completed in the time up to t
      self.t= self.t0  # the t before self.times[0].t
      self.min_width_for_next_message= 0
      self.n_done= 0
      self.n_error= 0
      self.n_no_overwrite= 0

   def wait_for_one(self, futures):
      done, not_done = concurrent.futures.wait(futures, return_when=concurrent.futures.FIRST_COMPLETED)
      futures[:]= not_done
      n= len(done)
      self.n_done += n
      self.times.append((time.time(), n))
      path_done= None
      for future in done:
         try:
            path_done= future.result()
         except Exception as e:
            self.n_error += 1
            if isinstance(e, RuntimeError):
               if isinstance(e, CannotOverwriteError):
                  self.n_no_overwrite += 1
               msg= str(e)
            else:
               msg= "".join(traceback.format_exception(type(e), e, e.__traceback__))
            self.message(msg)
      if path_done is None:
         return
      t= self.times[-1][0]
      while len(self.times)>1 and t - self.t > 5:
         self.t= self.times.pop(0)[0]
      average_speed= sum(n for t,n in self.times) / (t - self.t)
      eta= (self.n_files - self.n_done) / average_speed / 60
      message= "{}/{} {}% {}/s ETA:{}min {}".format(
              self.n_done,
              self.n_files,
              round(self.n_done*100/self.n_files),
              round(average_speed,1),
              round(eta),
              path_done)
      self.message(message, overwriteable=True)

   def message(self, msg, overwriteable=False):
      msg= msg.split("\n")
      l= len(msg[0])
      if l<self.min_width_for_next_message:
         msg[0] += ' '*(self.min_width_for_next_message - l)
      self.min_width_for_next_message= len(msg[-1]) if overwriteable else 0
      print("\n".join(msg), end= "\r" if overwriteable else "\n")

   def give_summary(self):
      msg= "Restored {}/{} files in {}min, {} errors, {} could not be overwritten".format(
         self.n_done - self.n_error,
         self.n_files,
         round((time.time()-self.t0)/60),
         self.n_error - self.n_no_overwrite,
         self.n_no_overwrite,
         )
      self.message(msg)



# -*- coding: UTF-8 -*-
# Copyright © 2017 Cornelius Grotjahn - This file is covered by the LICENCE file in the project root directory.
from   contextlib import contextmanager
import logging
import os.path
import queue
import re
import shutil
import sqlite3
import sys
import time
from   typing import Iterable, Tuple

from abus.cornelib import schapp_upgrade
from abus.rebuild import find_index_files, find_compressed_backups

class connect:
   def __init__(self, database_path, archive_root_path, allow_create=False):
      """
      Returns an object for the index database.

      :param database_path:
      :param allow_create: whether blank database should be created if `database_path` does not exist.
      :rtype: Database
      """
      self.db= Database(database_path, archive_root_path, allow_create)
   def __enter__(self):
      return self.db
   def __exit__(self, exc_type, exc_val, exc_tb):
      self.db.close()

class Database(object):
   def __init__(self, database_path, archive_root_path, allow_create):
      self.dbpath= database_path
      existed= os.path.exists(database_path)
      if not existed and not allow_create:
         raise RuntimeError("could not find database "+database_path)
      self.connection_pool_size= 0
      self.connection_pool= queue.Queue()
      self._set_connection_pool_size(1)
      self._check_schema(archive_root_path)
   def close(self):
      self._set_connection_pool_size(0)

   def _set_connection_pool_size(self, n):
      while self.connection_pool_size < n:
         conn= sqlite3.connect(self.dbpath, timeout=60, check_same_thread=False)
         conn.isolation_level= None # autocommit
         conn.execute("PRAGMA synchronous=OFF")
         self.connection_pool.put(conn)
         self.connection_pool_size += 1
      while self.connection_pool_size > n:
         self.connection_pool.get().close()
         self.connection_pool_size -= 1

   @contextmanager
   def _get_connection(self):
      """
      Returns a connection object from the pool with context manager.

      :rtype:sqlite3.Connection
      """
      conn= self.connection_pool.get()
      try:
         yield conn
      finally:
         self.connection_pool.put(conn)

   def _check_schema(self, archive_root_path):
      abus_src_dir= os.path.dirname(__file__)
      schema_file_name_pattern= re.compile(r"schema-([0-9]+)[.]sql")
      matches= ((schema_file_name_pattern.match(direntry.name), direntry.path)
                for direntry in os.scandir(abus_src_dir))
      target_schema, schema_file_path = max((int(m.group(1)),p) for m,p in matches if m)
      abus_version_pattern= re.compile(r"v([0-9]+)$")
      with open(schema_file_path) as stream:
         first_line= stream.readline()
         abus_version= abus_version_pattern.search(first_line).group(1)
      target_version= "{}.{}".format(target_schema, abus_version)

      with self._get_connection() as conn:
         try:
            current_version= conn.execute("select current_version from _database_version").fetchall()[0][0]
         except sqlite3.OperationalError as e:
            # did not have version table in versions <= 2.4, finding out manually which one it is
            tables= set(conn.execute("select name from sqlite_master where type='table'").fetchall())
            if ("completed_run",) in tables:
               current_version= "2"
            elif ("run",) in tables:
               current_version= "1"
            else:
               current_version= "0"
      current_schema= int(current_version.split(".")[0])
      if current_version == target_version:
         return
      if current_schema > target_schema:
         raise RuntimeError("ABuS {} is an older version than index database {}".format(abus_version, current_version))
      if 0 < current_schema < target_schema:
         upgrade_script_path= "{}/upgrade-{}.sql".format(abus_src_dir, current_schema)
      else:
         upgrade_script_path= None

      logging.info("index database needs upgrading")
      print("index database needs upgrading - this may take a while", file=sys.stderr)
      with self._take_backup():
         with self._get_connection() as conn:
            schapp_upgrade(conn, target_version, schema_file_path, upgrade_script_path)
            if 0 < current_schema < 3:
               # need to fill-in new columns from archive directory scan
               logging.info("reconstructing new index data")
               print("reconstructing new index data", file=sys.stderr)
               data= ((archive_dir, run_name) for run_name,archive_dir in find_index_files(archive_root_path))
               conn.executemany("update run set archive_dir= ? where run_name = ?", data)
               data= ((checksum,) for checksum in find_compressed_backups(archive_root_path))
               conn.executemany("update location set is_compressed= 1 where checksum = ?", data)

   @contextmanager
   def _take_backup(self):
      """Context manager that creates a backup copy of the database and
      restores or deletes it on exit
      depending on whether an exception has been raised"""
      connection_pool_size= self.connection_pool_size
      self._set_connection_pool_size(0)
      backup= self.dbpath + ".backup"
      logging.info("taking backup of index database")
      shutil.copyfile(self.dbpath, backup)
      self._set_connection_pool_size(1)
      try:
         yield
      except:
         self._set_connection_pool_size(0)
         logging.info("index database upgrade failed, restoring backup")
         shutil.copyfile(backup, self.dbpath)
         os.unlink(backup)
         raise
      else:
         os.unlink(backup)
         self._set_connection_pool_size(connection_pool_size)

   def _make_runname(self, timestamp):
      tm= time.localtime(timestamp)
      # format chosen to make the run name a "word" in most editors:
      return time.strftime("%Y_%m_%d_%H%M", tm)

   def open_backup_run(self, archive_dir):
      run_name= self._make_runname(time.time())
      with self._get_connection() as conn:
         n= conn.execute("select count(*) from run where run_name=?", [run_name]).fetchall()[0][0]
      if n>0:
         raise RuntimeError("There is already a run with the current timestamp (minute granularity")
      conn.execute("insert into run(run_name, is_complete, archive_dir) values(?,0,?)", [run_name, archive_dir])
      return run_name

   def complete_backup_run(self, run_name):
      with self._get_connection() as conn:
         cur= conn.execute("update run set is_complete= 1 where run_name = ?", [run_name])
         if cur.rowcount!=1:
            raise RuntimeError("could not complete run "+run_name)

   def get_archivedir_usage(self):
      with self._get_connection() as conn:
         rows= conn.execute("select archive_dir, count(*) as n from location group by archive_dir")
         counts= {archive_dir:n for archive_dir,n in rows}
         rows= conn.execute("select archive_dir, count(*) as n from run group by archive_dir")
         for archive_dir, n in rows:
            counts[archive_dir]= counts.get(archive_dir, 0) + n
         return counts

   def lookup_checksum(self, st_dev, st_ino, timestamp):
      if st_dev>=2**63: st_dev -= 2**64 # sqlite cannot handle u64
      if st_ino>=2**63: st_ino -= 2**64
      with self._get_connection() as conn:
         for (checksum,) in conn.execute("""
               select checksum
                  from checksum_cache
                  where dev=? and ino=? and timestamp=?""", (st_dev, st_ino, timestamp)):
            return checksum
         return None

   def remember_checksum(self, st_dev, st_ino, timestamp, checksum):
      if st_dev>=2**63: st_dev -= 2**64 # sqlite cannot handle u64
      if st_ino>=2**63: st_ino -= 2**64
      with self._get_connection() as conn:
         cur= conn.execute("""
            update checksum_cache
            set timestamp= ?, checksum= ?
            where dev=? and ino=?""", (timestamp, checksum, st_dev, st_ino))
         if cur.rowcount==0:
            conn.execute("insert into checksum_cache(timestamp, checksum, dev, ino) values(?,?,?,?)",
               (timestamp, checksum, st_dev, st_ino))

   def have_checksum(self, checksum: str) -> bool:
      """whether there is an existing backup for the given checksum"""
      with self._get_connection() as conn:
         rs= conn.execute("select archive_dir from location where checksum=?", [checksum]).fetchall()
      return len(rs)==1

   def remember_archivedir(self, checksum, archive_dir_rel, is_compressed):
      with self._get_connection() as conn:
         conn.execute("insert into location(checksum, archive_dir, is_compressed) values(?,?,?)",
                      (checksum, archive_dir_rel, is_compressed))

   def add_backup_entry(self, run_name, path, timestamp, checksum):
      with self._get_connection() as conn:
         conn.execute("""insert into content(run_name, path, timestamp, checksum)
            values(?,?,?,?)""", (run_name, path, timestamp, checksum))

   def get_all_backup_files(self):
      """
      Returns list of all checksums in backup

      :rtype: ((checksum, archive_dir, is_compressed))
      """
      with self._get_connection() as conn:
         return conn.execute("select checksum, archive_dir, is_compressed from location")

   def get_all_runs(self):
      """
      Returns list of all runs in DB

      :rtype: ((run_name, archive_dir))
      """
      with self._get_connection() as conn:
         return conn.execute("select run_name, archive_dir from run")

   def get_archive_contents(self, patterns, cutoff_date, show_all: bool) -> Iterable[Tuple[str,float,str,str,int]]:
      """
      Returns content of archive for listing or restore.

      :param patterns: glob-operator patterns for path to match any of
      :type patterns: string list
      :param cutoff_date: time from which files are ignored or None
      :type cutoff_date: time.time() format
      :param show_all: whether all files should be returned rather than just those from the latest run
      :returns: (path, timestamp, archive_dir, checksum, is_compressed)s
      """
      with self._get_connection() as conn:
         sql= "select distinct path, min(timestamp), archive_dir, content.checksum, is_compressed from content"
         sql_args= []
         where_clauses= [] # to be built now but added later
         where_args= []

         if cutoff_date is None:
            if show_all:
               # show all without cutoff date is just everything
               pass
            else:
               # show one without cutoff means latest run
               where_clauses.append("run_name = (select max(run_name) from content)")
         else:
            if show_all:
               # show all with cutoff date
               where_clauses.append("timestamp <= ?")
               where_args.append(cutoff_date)
            else:
               # show one with cutoff date requires subquery for the latest
               sql += " join (select path as p, max(timestamp) as t from content "
               sql +=       " where timestamp <= ?"
               sql_args.append(cutoff_date)
               sql +=       " group by path) as latest"
               sql +=  " on path = latest.p and timestamp = latest.t"

         sql += " join location on location.checksum = content.checksum"

         if patterns:
            where_clauses.append("("
               +" or ".join("path glob ?" for x in patterns)
               +")")
            where_args.extend(patterns)

         if where_clauses:
            sql += " where " + " and ".join(where_clauses)
            sql_args.extend(where_args)

         sql += " group by path, archive_dir, content.checksum"
         sql += " order by path, min(timestamp) desc"
         yield from conn.execute(sql, sql_args)

   def rebuild_location_table(self, actual_content_list: Iterable[Tuple[str, str, bool]]) -> Tuple[int, int, int]:
      """
      Adjusts the location table to reflect data in dictionary.

      :param actual_content_list: (checksum, archive_dir, is_compressed)s of required state
      :type actual_content_list: iterable
      :returns: counts: (updates, inserts, deletes)
      """
      actual_content= {checksum:(archive_dir,is_compressed)
                       for checksum,archive_dir,is_compressed in actual_content_list}
      updates= []
      deletes= []
      with self._get_connection() as conn:
         for checksum,archive_dir,is_compressed in conn.execute("select checksum, archive_dir, is_compressed from location"):
            # leaving actual_content with all records that are not in database:
            actual_values= actual_content.get(checksum)
            if actual_values is None:
               deletes.append((checksum,))
            else:
               del actual_content[checksum]
               if actual_values!=(archive_dir,is_compressed!=0):
                  archive_dir, is_compressed= actual_values
                  updates.append((archive_dir, is_compressed, checksum))
         conn.executemany("delete from location where checksum = ?", deletes)
         conn.executemany("update location set archive_dir= ?, is_compressed= ? where checksum = ?", updates)
         data= ((k,a,b) for k,(a,b) in actual_content.items())
         conn.executemany("insert into location(checksum,archive_dir,is_compressed) values(?,?,?)", data)
      return len(updates), len(actual_content), len(deletes)

   def rebuild_content(self, run_name, run_archive_dir, checksum_timestamp_path_rows):
      """
      Replaces all rows in content table for a given run_name with the given values.

      :param run_name:Run whose content rows will all be deleted and then replaced.
      :type run_name: str
      :param checksum_timestamp_path_rows: New rows (without run_nam column itself)
      :type checksum_timestamp_path_rows: (str,str,str)
      :return: number of records changed, added, deleted
      :rtype: tuple
      """
      typed_ctp= ((c,float(t),p) for c,t,p in checksum_timestamp_path_rows)
      with self._get_connection() as conn:
         conn.execute("""create temp table if not exists required_content(
            path text not null primary key,
            timestamp float not null,
            checksum text not null)""")
         conn.execute("delete from required_content")
         conn.executemany("insert into required_content(checksum, timestamp, path) values(?,?,?)", typed_ctp)
         cur= conn.execute("delete from required_content where checksum='error'")
         run_is_complete= cur.rowcount==0
         changed= conn.execute("""select count(*)
            from content join required_content on content.path=required_content.path
               and (content.checksum!=required_content.checksum or content.timestamp!=required_content.timestamp)
            where content.run_name=?""", [run_name]).fetchall()[0][0]
         new= conn.execute("""select count(*)
            from required_content left join content on content.path=required_content.path and content.run_name=?
            where content.path is null""", [run_name]).fetchall()[0][0]
         removed= conn.execute("""select count(*)
            from content left join required_content on required_content.path = content.path
            where content.run_name=? and required_content.path is null""", [run_name]).fetchall()[0][0]
         conn.execute("delete from content where run_name=?", [run_name])
         conn.execute("""insert into content(run_name, path, timestamp, checksum)
            select ?, path, timestamp, checksum from required_content""", [run_name])
         exists= conn.execute("select archive_dir, is_complete from run where run_name=?", [run_name]).fetchall()
         if len(exists)==0:
            new += 1
            conn.execute("""insert into run(run_name, archive_dir, is_complete) values(?,?,?)""",
                         (run_name, run_archive_dir, run_is_complete))
         elif exists[0]!=(run_archive_dir, int(run_is_complete)):
            changed += 1
            conn.execute("""update run set archive_dir= ?, is_complete= ? where run_name=?""",
                         (run_archive_dir, run_is_complete, run_name))
      return changed, new, removed

   def remove_runs(self, other_than: Iterable[str]):
      """
      Removes all runs from content and completed_runs, whose run_name is not in `other_than`

      :return: number of deleted rows
      """
      other_than= list(other_than)
      place_holders= ",?" * len(other_than)
      n= 0
      with self._get_connection() as conn:
         for table in "run", "content":
            stmt= "delete from "+table+" where run_name not in("+place_holders[1:]+")"
            cur= conn.execute(stmt, other_than)
            n += cur.rowcount
      return n

   def remove_location_entry(self, checksum):
      """
      Reflects in DB that a backup file has been deleted

      :param checksum: of deleted location file
      """
      with self._get_connection() as conn:
         conn.execute("delete from location where checksum=?", [checksum])

   def get_purgeable_backups(self, rounders):
      """
      Returns list of backup files that are superfluous and to be deleted.

      :param: rounders: retention policy definition: (rounder,age)-list
      :return:  (checksum, archive_dir, path, timestamp)-list
      """
      case_clause= "case " +"when timestamp >= ? then round(timestamp/?)*? "*len(rounders) +"else 0 end"
      now= time.time()
      params= []
      for r,a in sorted(rounders):
         params.extend([now-a*86400, r*86400, r*86400])

      with self._get_connection() as conn:
         cur= conn.execute("""
            with SLOTTED as(
                  select distinct path, timestamp, """+case_clause+""" as slot
                  from content
                  )
               ,KEEP_VERSIONS as(
                  select path, max(timestamp) as timestamp
                  from SLOTTED
                  group by path, slot
                  )
               ,KEEP_CHECKSUMS as(
                  select distinct content.checksum
                  from content join KEEP_VERSIONS on content.path = KEEP_VERSIONS.path
                     and content.timestamp = KEEP_VERSIONS.timestamp
                  )
               ,PURGE as(
                  select distinct location.checksum, location.archive_dir, content.path, content.timestamp
                  from location
                     join content on content.checksum = location.checksum
                  where location.checksum not in(select checksum from KEEP_CHECKSUMS)
                  )
            select *
            from PURGE
            order by checksum, path, timestamp
            """, params)
         return cur.fetchall()

   def purge_runs(self, run_names: Iterable[str]) -> None:
      """
      Purges given runs after checking that they can be purged.
      """
      run_names= set(run_names)
      run_names.intersection_update(r for r,a in self.get_purgeable_runs())
      run_names= list(run_names)
      if run_names:
         place_holders= ",?" * len(run_names)
         with self._get_connection() as conn:
            for table in ("run", "content"):
               conn.execute("delete from {} where run_name in({})".format(table, place_holders[1:]), run_names)

   def get_purgeable_runs(self) -> Iterable[Tuple[str,str]]:
      """
      Returns runs that can be purged.

      :returns: (run_name, archive_dir)s
      """
      with self._get_connection() as conn:
         return conn.execute("""
            select run_name, archive_dir
            from run
            where run_name not in(select run_name
               from content join location on location.checksum = content.checksum)""")
